//============================================================================
//
//                              Program/Module
//                                   from
//                C++ package for verified numerical integration
//                  of singular and strong singular functions
//                      Copyright (c) 2000 Stefan Wedner
//
// This program/module is free software for non-commercial use.
//
//
//============================================================================
//----------------------------------------------------------------------------
//
// File:     cubature.h
// See also: cubature.C
// Purpose:  header (see cubature.C)
//     
// ===========================================================================


#include"cubature.h"
#include<fstream>
#include<rmath.hpp>


#ifdef  TIME_CHECK
#include<time.h> 
  clock_t  start_taylor, end_taylor, start_function, end_function;
#endif   

// ----------------------------------------------------------------------------
// --- constants and static elements-------------------------------------------
// ----------------------------------------------------------------------------

dotprecision base_interval::r_dot;


// --- class xriemann ---------------------------------------------------------

const int xriemann::n_min    =   8;
const int xriemann::n_max    =  17;
const int xriemann::n_stride =   3;
const int xriemann::n_dim    =  xriemann::n2n_index(n_max);
const int xriemann::m_min    =   7;
const int xriemann::m_max    =  16;
const int xriemann::m_stride =   3;
const int xriemann::m_dim    =  xriemann::m2m_index(m_max);


interval ** xriemann::node;
interval ** xriemann::weight;
int      *  xriemann::peano_n;
interval ** xriemann::peano    =  xriemann::init_data();

// --- class xcauchy ----------------------------------------------------------

const int xcauchy::m_min        =  riemann::m_min + 1;
const int xcauchy::m_max        =  riemann::m_max + 1;
const int xcauchy::m_stride     =  riemann::m_stride;
const int xcauchy::m_dim        =  riemann::m_dim;


interval** xcauchy::peano; 
int*       xcauchy::peano_n;
interval*  xcauchy::peano_w;

// --- class integral ---------------------------------------------------------

const int  integral::max_heapsize     = 500;
real       integral::epsilon;
int        integral::integral_counter = 0;

// --- class cauchy_integral --------------------------------------------------

int cauchyxriemann_integral::cauchy_counter = 0;



// ----------------------------------------------------------------------------
// --- auxiliary functions read_data, read_datasymm ---------------------------
// ----------------------------------------------------------------------------



// --- read_datasymm (nodes and weights ) -------------------------------------


void read_datasymm(char* name, interval** node, interval** weight, 
                   int n_min = 1, int n_stride = 1){

  int  i, n;
  real a_down, a_up;
  interval *pointer1, *pointer2;
  ifstream external_data(name);
  if(! external_data) cout << " error! : file " << name 
                           << " not available! " << endl;
  external_data >> n;

  pointer1   = node[(n - n_min) / n_stride];
  pointer2   = weight[(n - n_min) / n_stride];

  for(i=0; i<= n / 2 - 1; i++){
    external_data >> Hex >>  a_down;
    external_data >> Hex >>  a_up;
    pointer1[i] = interval(a_down,a_up);
    pointer1[n - i - 1] = -  pointer1[i]; 
  }
  if(n%2 ){
    external_data >> Hex >> a_down;
    external_data >> Hex >> a_up;
    pointer1[n/2] = interval(a_down,a_up);
  }
  for(i=0; i<= n / 2 - 1; i++){
    external_data >> Hex >>  a_down;
    external_data >> Hex >>  a_up;
    pointer2[i] = interval(a_down,a_up);
    pointer2[n - i - 1] =   pointer2[i]; 
  }
  if(n%2 ){
    external_data >> Hex >> a_down;
    external_data >> Hex >> a_up;
    pointer2[n/2] = interval(a_down,a_up);
  }  
  cout << Dec;
  external_data.close();
} 



// --- read_data (nodes and weights not symmetric) ----------------------------


void read_data(char* name, interval** node, interval** weight, 
               int n_min = 1, int n_stride = 1){
  int  i, n;
  real a_down, a_up;
  interval* pointer1, *pointer2;
  ifstream external_data(name);

  if(! external_data) cout << " error! : file " << name 
                           << " not available! " << endl;
  external_data >> n;

  pointer1   = node[(n - n_min) / n_stride];
  pointer2   = weight[(n - n_min) / n_stride];

  for(i=0; i< n; i++){
    external_data >> Hex >>  a_down;
    external_data >> Hex >>  a_up;
    pointer1[i] = interval(a_down,a_up);
  }
  for(i=0; i< n; i++){
    external_data >> Hex >>  a_down;
    external_data >> Hex >>  a_up;
    pointer2[i] = interval(a_down,a_up);
  }
  cout << Dec;
  external_data.close();
} 



// --- read_data (for peano constants) ----------------------------------------


int read_data(char* name, interval** peano, int* peano_n, int m_max, 
             int m_min = 1, int m_stride = 1, int n_min = 1, int n_stride = 1){

  int  i, n, data_min, data_max, datai_max;
  real a_down, a_up;
  ifstream external_data(name);
  if(! external_data) cout << " error! : file " << name 
                           << " not available! " << endl;
  
  external_data >> n;
  if (n < n_min) throw ;          //  error wrong n  
  
  external_data >> data_min;
  external_data >> data_max;
  
  for(i= data_min; i< m_min; i++){
    external_data >> Hex >>  a_down;
    external_data >> Hex >>  a_up;
    external_data >> Hex >>  a_down;
    external_data >> Hex >>  a_up;
  }

  external_data >> Hex >>  a_down;
  external_data >> Hex >>  a_up;
  external_data >> Hex >>  a_down;
  external_data >> Hex >>  a_up;

  n = (n - n_min) / n_stride;
  if (data_max > m_max) datai_max = (m_max - m_min) / m_stride;
  else datai_max = (data_max - m_min) / m_stride;
 
  peano[0][n]  = interval(a_down,a_up);

  for(i= 1; i<= datai_max; i++){
    for( int j=0; j < m_stride; j++){
      external_data >> Hex >>  a_down;
      external_data >> Hex >>  a_up;
      external_data >> Hex >>  a_down;
      external_data >> Hex >>  a_up;
    }
    peano[i][n]  = interval(a_down,a_up);
  }
  for(i; i<= (m_max - m_min) / m_stride; i++)
    peano[i][n] = 0.0;
  external_data.close();
  cout << Dec;
  return 0;
}


// ----------------------------------------------------------------------------
// --- class integral ---------------------------------------------------------
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
// --- constructors class integral --------------------------------------------
// ----------------------------------------------------------------------------

// --- integral::integral------------------------------------------------------
// --- without arguments (important for derived classes) ----------------------


integral::integral() :f(xriemann::m_max, xriemann::m_max), 
            ftransposed(xriemann::m_max, xriemann::m_max) {

  integral_counter++;
  heapsize          = 0;
  heap              = new pbase_interval[max_heapsize];
  fcounter          = 0;
  divided_intervals = 0;
  Rinf              = 0;
  Rsup              = 0;
  Approximation     = 0;
  eps_rel           = 0;
  eps_abs           = 0;
}


// --- integral::integral------------------------------------------------------


integral::integral(integrand& g):f(xriemann::m_max, xriemann::m_max), 
                               ftransposed(xriemann::m_max, xriemann::m_max) {

  integral_counter++;
  heapsize          = 0;
  heap              = new pbase_interval[max_heapsize];
  fcounter          = 0;
  divided_intervals = 0;
  Rinf              = 0;
  Rsup              = 0;
  Approximation     = 0;
  f                 = g;
  ftransposed       = f.transpose();
  eps_rel           = 0;
  eps_abs           = 0;
}



// --- destruktor class integral ----------------------------------------------
// --- free memory for heap and nodes, weights and peanoconstants -------------


integral::~integral(){ 

  heapsize = 0;
  delete[] heap ;
  integral_counter--;
  //  if (!integral_counter)  xriemann::free_data(); 

}



// --- integral::start_division -----------------------------------------------


void integral::start_division(){

  int m_index = xriemann::m_dim,
      n_index = xriemann::n_dim;

  riemann* start = new riemann(xinf, xsup, yinf, ysup, m_index, n_index, 
                               m_index, n_index,  *this);
  start->choose_xcf(0.5 * eps_abs);
  start->choose_ycf(0.5 * eps_abs);
  push( *start);

}



// --- integral::integrate ----------------------------------------------------
// --- using absolute error ---------------------------------------------------


interval integral::integrate(real a, real b, real c, real d, real epsabs){

  xinf = a; xsup = b;
  yinf = c; ysup = d;
  eps_abs = epsabs;
  epsilon = 2 *  eps_abs / ( (xsup -xinf) * (ysup -yinf) );
  Approximation = 0;
  start_division();
  gaa();
 
  real tmp_inf, tmp_sup;
  interval result;

  rnd(Rinf, tmp_inf, RND_DOWN);
  rnd(Rsup, tmp_sup, RND_UP );
  rnd(Approximation, result);
  return result + interval(tmp_inf, tmp_sup);    
}



interval integral::integrate(interval a, interval b, interval c, 
                                                     interval d, real epsabs){
  integral& ri = static_cast<integral&>(*this);
  xriemann* tmpr;
  int  m_index = xriemann::m_dim,
       n_index = xriemann::n_dim;
 
  eps_abs = epsabs;
  epsilon = 2 * eps_abs / ( (Sup(b) -Inf(a)) * (Sup(d) - Inf(c)) );
  Approximation = 0;

  xinf = Sup(a); xsup = Inf(b);
  yinf = Sup(c); ysup = Inf(d);
  Approximation = 0;

   if (Inf(a) != Sup(a)){
    tmpr = new xriemann(Sup(c), Inf(d), Inf(a), Sup(a), m_index, n_index,  
                              ri , epsilon, true);
    if (Inf(b) != Sup(b)){
        tmpr = new xriemann(Sup(c), Inf(d), Inf(b), Sup(b), m_index, n_index, 
                                                        ri , epsilon, true); 
        if (Inf(c) != Sup(c)){
          tmpr = new xriemann(Inf(a),Sup(a),Inf(c),Sup(c), ri);
          tmpr = new xriemann(Inf(b),Sup(b),Inf(c),Sup(c), ri);
          tmpr = new xriemann(Sup(a),Inf(b),Inf(c),Sup(c), 
                              m_index, n_index, ri, epsilon);
	}
        if (Inf(d) != Sup(d)){
          tmpr = new xriemann(Inf(a),Sup(a),Inf(d),Sup(d), ri);
          tmpr = new xriemann(Inf(b),Sup(b),Inf(d),Sup(d), ri);
          tmpr = new xriemann(Sup(a),Inf(b),Inf(d),Sup(d), 
                              m_index, n_index, ri, epsilon);
	}      
    }
    else{
        if (Inf(c) != Sup(c)){
          tmpr = new xriemann(Inf(a),Sup(a),Inf(c),Sup(c), ri);
          tmpr = new xriemann(Sup(a),Inf(b),Inf(c),Sup(c), 
                              m_index, n_index, ri, epsilon); 
	}
        if (Inf(d) != Sup(d)){
          tmpr = new xriemann(Inf(a),Sup(a),Inf(d),Sup(d),ri);
          tmpr = new xriemann(Sup(a),Inf(b),Inf(d),Sup(d), 
                              m_index, n_index, ri, epsilon);
	}      
    }
  }
  else{
    if (Inf(b) != Sup(b)){
        tmpr = new xriemann(Sup(c), Inf(d), Inf(b), Sup(b), m_index, n_index, 
                                                        ri , epsilon, true); 
        if (Inf(c) != Sup(c)){
          tmpr = new xriemann(Inf(b),Sup(b),Inf(c),Sup(c),ri);
          tmpr = new xriemann(Sup(a),Inf(b),Inf(c),Sup(c), 
                                       m_index, n_index, ri, epsilon);
	}
        if (Inf(d) != Sup(d)){
          tmpr = new xriemann(Inf(b),Sup(b),Inf(d),Sup(d),ri);
          tmpr = new xriemann(Sup(a),Inf(b),Inf(d),Sup(d), 
                                       m_index, n_index, ri, epsilon);
	}      
    }
    else{
        if (Inf(c) != Sup(c)) tmpr = new xriemann(Sup(a), Inf(b), Inf(c), 
                                     Sup(c), m_index, n_index, *this, epsilon);
        if (Inf(d) != Sup(d)) tmpr = new xriemann(Sup(a), Inf(b), Inf(d),
                                     Sup(d), m_index, n_index, ri, epsilon);
	}      
  }
  start_division();
  gaa();
 
  real tmp_inf, tmp_sup;
  interval result;

  rnd(Rinf, tmp_inf, RND_DOWN);
  rnd(Rsup, tmp_sup, RND_UP );
  rnd(Approximation, result);
  return result + interval(tmp_inf, tmp_sup);    

}


// --- integral::integrate ----------------------------------------------------
// --- improve absolute error with the same integration range -----------------


interval integral::integrate(real epsabs){

  Approximation = 0;
  gaa();

  real tmp_inf, tmp_sup;
  interval result;

  rnd(Rinf, tmp_inf, RND_DOWN);
  rnd(Rsup, tmp_sup, RND_UP );
  rnd(Approximation, result);
  return result + interval(tmp_inf, tmp_sup);  
}



// --- integral::gaa ----------------------------------------------------------
// global adaptive algorithm: subdivide range with biggest  remainder diam.


void  integral::gaa(){

  base_interval* old_range;

#ifdef TIME_CHECK
  start_taylor = clock();
#endif 

  while( (Rsup - Rinf) > eps_abs && heapsize < max_heapsize ){
    old_range = pop();
    divided_intervals++; 
    old_range->subdivide();
  }

    
#ifdef TIME_CHECK
  end_taylor = clock();
  start_function = clock();
#endif 
  // compute approximation

  for(int i=0; i<heapsize; i++){
     old_range = heap[i].pointer;
     if (old_range->approx) 
        accumulate(Approximation, 1.0, old_range->approximation);
     else old_range->add_approxsum();
     fcounter += old_range->xn * old_range->yn;   
  }
  
#ifdef TIME_CHECK
  end_function = clock();
  cout <<  start_taylor << " , " << end_taylor << endl;
  cout <<  start_function << " , " << end_function << endl;
#endif 
  if (heapsize == max_heapsize){ cout << " HEAP" << endl; throw ;}
    // max heap_size reached! 
}



// ----------------------------------------------------------------------------
// --- integral: input output -------------------------------------------------
// ----------------------------------------------------------------------------
// --- operator << ------------------------------------------------------------


ostream& operator<<(ostream& os, const integral& out){
  interval Q, R;
  real rinf, rsup;
  rnd(out.Approximation, Q); rnd(out.Rinf, rinf, RND_DOWN); 
  rnd(out.Rsup, rsup, RND_UP);
  R =  interval(rinf, rsup);
  //os << " function: " << out.f << endl;
  //os << " integrationbounds:   " << out.xinf << " , " << out.xsup 
  //   << " x " << out.yinf << " , " << out.ysup << endl;
  os << " number of intervals : " << out.heapsize << " " 
     << "(" << out.divided_intervals << ")" << endl;
  os << " #f                  : " << out.fcounter << endl << endl; 
  os << " approximationsum    : " << Q << endl 
     << " d(approximationsum) : " << diam(Q) << endl << endl;
  os << " remainder           : " << R << endl
     << " d(remainder)        : " << diam(R) << endl << endl;
  Q += R;
  os << " enclosure           : " << Q << endl
     << " d(enclosure)        : " << diam(Q) << endl;
}

// output tex for diss
// ostream& operator<<(ostream& os, const integral& out){
//    interval Q, R;
//    real rinf, rsup;
//    rnd(out.Approximation, Q); rnd(out.Rinf, rinf, RND_DOWN); 
//    rnd(out.Rsup, rsup, RND_UP);
//    R =  interval(rinf, rsup);  
//    Q += R;
//    os << " &  " << Q << " & " << diam(Q) << " & " << endl;
//    os << " " <<  out.heapsize << " & " <<  out.divided_intervals << " & ";
//    os << out.fcounter << " \\\ "; 
// }


// --- integral::print_division -----------------------------------------------


void integral::print_division(){

  base_interval* old_range;

  for(int i=0; i<heapsize; i++){
     old_range = heap[i].pointer;
     old_range->print_interval(cout);
  }
}


void integral::print_gnuplot(){

  base_interval* old_range;

  for(int i=0; i<heapsize; i++){
     old_range = heap[i].pointer;
     if (!old_range->transposed)
       cout << 1 << ", " << 1 << ", " << old_range->xinf << ", " << old_range->xsup 
            << ","  << old_range->yinf << "," << old_range->ysup << endl;
     else
       cout << 1 << ", " << 1 << ", " << old_range->yinf << ", " << old_range->ysup 
            << ", "<< old_range->xinf << ", " << old_range->xsup << endl;
  }
}


// ----------------------------------------------------------------------------
// --- class base_interval ----------------------------------------------------
// ----------------------------------------------------------------------------

// --- operator << ------------------------------------------------------------

 
ostream& operator<<(ostream& os, const base_interval& out){

  os  << endl << " R " << out.r_inf  << " ,"  << out.r_sup 
      << " diam: " << out.dremainder << endl; 
   if (!out.transposed){
     os << " [" << out.xinf << " , " << out.xsup << " ]" << " x ";
     os << " [" << out.yinf << " , " << out.ysup << " ]" << endl;
     os << " xm: " << out.xm << " , xn: " << out.xn << endl;
     os  << " ym: " << out.ym << " , yn: " << out.yn << endl;
   }
   else{
     os << " [" << out.yinf << " , " << out.ysup << " ]" << " x ";
     os << " [" << out.xinf << " , " << out.xsup << " ]" << endl;
     os << " xm: " << out.ym << " , xn: " << out.yn << endl;
     os  << " ym: " << out.xm << " ,yn: " << out.xn << endl;
   }

   return os;
}



// ----------------------------------------------------------------------------
// --- class pbase_interval----------------------------------------------------
// ----------------------------------------------------------------------------
// (pointer layer for base_interval to speed up heapsort)
// ----------------------------------------------------------------------------


ostream& operator<<(ostream& os, const pbase_interval& out){
  os << out.pointer->dremainder;
  return os;
}

istream& operator>>(istream& is, pbase_interval& in){
  is >> in.pointer->dremainder;
  return is;
}

bool  pbase_interval::operator<(const pbase_interval& t) const{
  return pointer->dremainder < t.pointer->dremainder;
}

bool  pbase_interval::operator>(const pbase_interval& t) const{
  return pointer->dremainder > t.pointer->dremainder;
}

bool  pbase_interval::operator==(const pbase_interval& t) const{
  return pointer->dremainder == t.pointer->dremainder;
}


// ----------------------------------------------------------------------------
// --- class xriemann ---------------------------------------------------------
// ----------------------------------------------------------------------------


// --- xriemann::add_approxsum ------------------------------------------------
// 
// computes enclosure of approximation sum for
//                 int_xinf^xsup int_[yinf,ysup]^ysup or
//                 int_xinf^xsup int_yinf^[yinf,ysup] 
// if transposed == true
// computes enclosure of approximation sum for
//                 int_[xinf,xsup]^xsup int_yinf^ysup or
//                 int_xinf^[xinf,xsup] int_yinf^ysup 

void xriemann::add_approxsum(){ 

  integrand* fr = &ri.f; ;

  if (transposed) fr = &ri.ftransposed;

  interval  *x   = node[xn_index] , 
            *w   = weight[xn_index],
            m, m2, c;
 
  m  = (xsup - interval(xinf,xinf)) * 0.5;
  m2 = (ysup - interval(yinf,ysup)) * m;
  c  = (xsup + interval(xinf,xinf)) * 0.5;

  for(int i=0; i< xn; i++)
   accumulate(ri.Approximation, m2 * w[i] , 
              fr->value( m * x[i] + c, interval(yinf, ysup) ) ); 
}



// --- xriemann::subdivide ----------------------------------------------------


void xriemann::subdivide(){
  // eps_local = eps_abs * B'x / B
  // test removing 0.5
   real eps_local = ri.eps_abs * (xsup - xinf) / (ri.xsup - ri.xinf)
                    / ( ri.ysup - ri.xinf),  midpoint;

   r_dot = 0;
   accumulate(r_dot, 0.5,  xinf);
   accumulate(r_dot, 0.5,  xsup);
   rnd(r_dot, midpoint);
   xriemann* left = new xriemann(xinf, midpoint, yinf, ysup,
                                       xm_index, xn_index, ri, transposed);
   left->choose_qf(eps_local);
   ri.push(*left);
   xinf = midpoint; 
   choose_qf(eps_local);
   approx = false;
   ri.push(*this);
}


 
// --- xriemann::choose_qf ----------------------------------------------------

void xriemann::choose_qf(real eps_local){

   interval tk_old, tk_new;
   real  dr_old, dr_new, factor;  
   int  k, k_index;

   integrand*      fr = &ri.f;
   if (transposed) fr = &ri.ftransposed;

   if (xm < m_max){      // workaround: 
     fr->taylor(xm + m_stride, interval(xinf, xsup), interval(yinf, ysup));
     tk_old  = fr->taylor(xm);
   }
   else
     tk_old = fr->taylor(xm, interval(xinf, xsup), interval(yinf, ysup));

    factor  = power( (xsup - xinf) * 0.5, xm + 1 ) * diam(tk_old);

   // looking for appropriate n
 
   dr_old = Sup(peano[xm_index][n_dim]) * factor; 
   if (xm != 2 * xn) dr_old *= 2.0; 

   if (dr_old < eps_local){
    
     for(k_index = n_dim - 1, k = n_max - n_stride; 
         k_index >= peano_n[xm_index]; k_index--, k -= n_stride)
     {
        dr_new = Sup(peano[xm_index][k_index]) * factor;
        if (xm != 2 * k) dr_new *= 2.0;
        if(dr_new > eps_local) break;  
     }
     set_xriemann(tk_old, k_index + 1, k + n_stride);          
     return;    
   }

   // looking for appropriate m
   if (xm_index > 0){  
     k = xm - m_stride; k_index = xm_index -1;
     tk_new = power( (xsup - interval(xinf,xinf) ) * 0.5, k + 1) * 
                   fr->taylor(k);
     if ( 2 * diam(tk_new * peano[k_index][n_dim]) < dr_old){ 
       set_xriemann(k_index, k, tk_new);          
       return; 
     }
   }

   if(xm < m_max){ 
     k = xm + m_stride; k_index = xm_index + 1;
     tk_new = power( ( xsup - interval(xinf, xinf) ) * 0.5, k + 1) * 
       fr->taylor(k); 
    if (k != 2 * n_max) dr_old *= 0.5; 
    if (diam(tk_new * peano[k_index][n_dim])  < dr_old){ 
       set_xriemann(k_index, k, tk_new);  
       return; 
    }
   }
   set_xriemann(tk_old, xn_index, xn);
}



// --- xriemann::init_data ----------------------------------------------------


interval** xriemann::init_data(){
// init nodes, weights, peanoconstants
  int i;

  node    = new interval*[n_dim + 1];
  weight  = new interval*[n_dim + 1];
  peano   = new interval*[m_dim + 1];
  peano_n = new int[m_dim + 1];

  for(i = 0; i<= n_dim; i++){
    node[i]   = new interval[ n_index2n(i) ];
    weight[i] = new interval[ n_index2n(i) ];
  } 

  read_datasymm("./data/gaussleg8.hex" , node, weight, n_min, n_stride);
  read_datasymm("./data/gaussleg11.hex", node, weight, n_min, n_stride);
  read_datasymm("./data/gaussleg14.hex", node, weight, n_min, n_stride);
  read_datasymm("./data/gaussleg17.hex", node, weight, n_min, n_stride);
 
  for(i = 0; i <= m_dim; i++) 
    peano[i] = new interval[n_dim + 1];
 
  read_data("./data/peanoleg8.hex", peano, peano_n, m_max , m_min, m_stride,
                                                            n_min, n_stride);

  read_data("./data/peanoleg11.hex", peano, peano_n, m_max, m_min, m_stride,
                                                            n_min, n_stride);

  read_data("./data/peanoleg14.hex", peano, peano_n, m_max, m_min, m_stride,
                                                            n_min, n_stride);

  read_data("./data/peanoleg17.hex", peano, peano_n, m_max, m_min, m_stride,
                                                            n_min, n_stride);


  for(int j=0; j<=m_dim; j++){
    i = -1;
    do{ 
      if (i<n_dim) i++;
      else throw ;                 //  remove!!!!!         
    }
    while(peano[j][i] == interval(0,0));
    peano_n[j] = i;
  }
  return peano;
}



// --- xriemann::free_data ----------------------------------------------------


void xriemann::free_data(){

 int i;

  for(i = 0; i<= n_dim; i++){
    delete node[i];
    delete weight[i];
  } 
  delete[] node;
  delete[] weight;
  for(i = 0; i <= m_dim; i++) 
    delete[] peano[i];
  delete[] peano;
  delete[] peano_n;
}



// --- xriemann::set_xriemann -------------------------------------------------


inline void xriemann::set_xriemann(int k_i, int k, interval tk){
  interval r;
  xm_index   = k_i;
  xm         = k; 
  r          = (ysup - interval(yinf,ysup)) * peano[xm_index][xn_index] * tk;
  if (xm != 2 * xn) r -= r;
  r_inf      = Inf(r);
  r_sup      = Sup(r);
  dremainder = diam(r);
  return;

}


inline void xriemann::set_xriemann(interval tk, int k_i, int k){
 interval r;
  xn_index    = k_i;
  xn          = k; 
  r          = peano[xm_index][xn_index] * (ysup - interval(yinf,ysup)) *
               power((xsup- interval(xinf, xinf))* 0.5, xm + 1) * tk;
  if (xm != 2 * xn) r  -= r; 
  r_inf      = Inf(r);
  r_sup      = Sup(r);
  dremainder = diam(r);
  return;
}


// ----------------------------------------------------------------------------
// --- class riemann ----------------------------------------------------------
// ----------------------------------------------------------------------------

// --- riemann::add_approxsum -------------------------------------------------


void riemann::add_approxsum(){ 
 
  int i,j;
  interval  *x   = node[xn_index] ,  *y  = node[yn_index], 
            *xw  = weight[xn_index], *yw = weight[yn_index],
            xm, xc, ym, yc;

  xm  = (xsup - interval(xinf, xinf)) * 0.5;
  xc  = (xsup + interval(xinf, xinf)) * 0.5;
  ym  = (ysup - interval(yinf, yinf)) * 0.5;
  yc  = (ysup + interval(yinf, yinf)) * 0.5;

  for(i=0; i< xn; i++)
    for(j=0; j< yn; j++)
       accumulate(ri.Approximation, xm * ym * xw[i] * yw[j], 
              ri.f.value( xm * x[i] + xc,  ym * y[j] + yc ) ); 
  //  cout << " durchmesser Rie nachher " << diam(ri.Approximation) << endl;
}



// --- riemann::set_xriemann --------------------------------------------------


inline void riemann::set_xriemann(int k_i, int k, interval tk){

  interval r;
  xm_index    = k_i;
  xm          = k; 
  r          = peano[xm_index][xn_index] * tk;
  if (xm != 2 * xn) r -= r;
  r_inf      = Inf(r);
  r_sup      = Sup(r);
  dremainder = diam(r);
  return;

}



inline void riemann::set_xriemann(interval tk, int k_i, int k){

 interval r;
  xn_index    = k_i;
  xn          = k; 
  r          = peano[xm_index][xn_index] * 
               power((xsup- interval(xinf, xinf))* 0.5, xm + 1) * 
               (ysup - interval(yinf, yinf)) * tk;
  if (xm != 2 * xn) r  -= r; 
  r_inf      = Inf(r);
  r_sup      = Sup(r);
  dremainder = diam(r);
  return;
}



// --- riemann::set_yriemann --------------------------------------------------


inline void riemann::set_yriemann(int k_i, int k, interval tk){
  interval r;
  ym_index    = k_i;
  ym          = k; 
  r           = peano[ym_index][yn_index] * tk;
  if (ym != 2 * yn) r -= r;
  r_inf      += Inf(r);
  r_sup      += Sup(r);
  dremainder += diam(r);
  return;

}


inline void riemann::set_yriemann(interval tk, int k_i, int k){
 interval r;
  yn_index    = k_i;
  yn          = k; 
  r           = peano[ym_index][yn_index] * 
                power((ysup- interval(yinf, yinf))* 0.5, ym + 1) * 
                ( xsup - interval(xinf, xinf) )  * tk;
  if (ym != 2 * yn) r  -= r; 
  r_inf      += Inf(r);
  r_sup      += Sup(r);
  dremainder += diam(r);
  return;
}



// --- riemann::subdivide -----------------------------------------------------


void riemann::subdivide(){
  // vary eps_local 
  riemann* tmp;
  real xmidpoint, ymidpoint;
  real  eps_local =  ri.epsilon * (xsup - xinf) * (ysup -yinf);   
  if (xdivide){
      r_dot = 0;
      accumulate(r_dot, 0.5,  xinf);
      accumulate(r_dot, 0.5,  xsup);
      rnd(r_dot, xmidpoint);
    if (ydivide){
      r_dot = 0;
      accumulate(r_dot, 0.5,  yinf);
      accumulate(r_dot, 0.5,  ysup);
      rnd(r_dot, ymidpoint); 

      tmp = new riemann(xinf, xmidpoint, yinf, ymidpoint,xm_index, xn_index, 
       ym_index, yn_index, ri, xdivide, ydivide,eps_local * 0.25);
  
      tmp = new riemann(xinf, xmidpoint, ymidpoint, ysup, xm_index, xn_index, 
       ym_index, yn_index, ri, xdivide, ydivide, eps_local * 0.25);

      tmp = new riemann(xmidpoint, xsup, yinf, ymidpoint, xm_index, xn_index,
       ym_index, yn_index, ri, xdivide, ydivide, eps_local * 0.25);
 
      xinf =  xmidpoint;
      yinf =  ymidpoint;
      choose_xcf(eps_local * 0.25);
      choose_ycf(eps_local * 0.25);
      ri.push(*this);  
    }
    else{  // xdivide and ! ydivide
      tmp = new riemann(xinf, xmidpoint, yinf, ysup, xm_index, xn_index, 
       ym_index, yn_index, ri, xdivide, ydivide, eps_local * 0.5); 

      xinf =  xmidpoint;
      choose_xcf(eps_local * 0.5);
      choose_ycf(eps_local * 0.5);
      ri.push(*this);    
    }
  }
  else if(!xdivide){  // !xdivide
    r_dot = 0;
    accumulate(r_dot, 0.5,  yinf);
    accumulate(r_dot, 0.5,  ysup);
    rnd(r_dot, ymidpoint);

    tmp = new riemann(xinf, xsup, yinf, ymidpoint, xm_index, xn_index, 
       ym_index, yn_index, ri, xdivide, ydivide, eps_local * 0.5);

    yinf =  ymidpoint;
    choose_xcf(eps_local * 0.5);
    choose_ycf(eps_local * 0.5);
    ri.push(*this);    
  }
  else{  ri.push(*this);  xdivide = ydivide = true;}
}
 


// --- riemann::choose_xcf ----------------------------------------------------


void riemann::choose_xcf(real eps_local){

   interval tk_old, tk_new;
   real  dr_old, dr_new, factor;  
   int  k, k_index;

   if (xm < m_max && xdivide){      
     ri.f.taylor(xm + m_stride, interval(xinf, xsup), interval(yinf, ysup));
     tk_old  = ri.f.taylor(xm);
   }
   else 
     tk_old = ri.f.taylor(xm, interval(xinf, xsup), interval(yinf, ysup));

     factor = pow( (xsup - xinf) * 0.5, xm + 1 ) * (ysup - yinf) 
              * diam(tk_old);

   // looking for appropriate n
 
   dr_old = Sup(peano[xm_index][n_dim]) * factor; 
   if (xm != 2 * n_max) dr_old *= 2.0; 

   if (dr_old < eps_local){
    
     for(k_index = n_dim - 1, k = n_max - n_stride; 
         k_index >= peano_n[xm_index]; k_index--, k -= n_stride)
     {
        dr_new = Sup(peano[xm_index][k_index]) * factor;
        if (xm != 2 * k) dr_new *= 2.0;
        if(dr_new > eps_local) break;  
     }
     set_xriemann(tk_old, k_index + 1, k + n_stride);
     xdivide = false;          
     return;    
   }
   if (!xdivide) cout << "xdivide error ? " << endl;
   xdivide = true;
   // looking for appropriate m
   if (xm_index > 0){  
     k = xm - m_stride; k_index = xm_index -1;
     tk_new = power( (xsup - interval(xinf,xinf) ) * 0.5, k + 1) * 
              (ysup - interval(yinf, yinf)) * ri.f.taylor(k);
     if ( 2 * diam(tk_new * peano[k_index][n_dim]) <  dr_old){ 
       set_xriemann(k_index, k, tk_new);          
       return; 
     }
   }

   if(xm < m_max){ 
     k = xm + m_stride; k_index = xm_index + 1;
     tk_new = power( ( xsup - interval(xinf, xinf) ) * 0.5, k + 1) * 
       (ysup - interval(yinf, yinf)) * ri.f.taylor(k); 
     if (k != 2 * n_max) dr_old *= 0.5; 
     if (diam(tk_new * peano[k_index][n_dim])  < dr_old){ 
       set_xriemann(k_index, k, tk_new);  
       return; 
    }
   }
   set_xriemann(tk_old, xn_index, xn);
}



// --- riemann::choose_ycf ----------------------------------------------------


void riemann::choose_ycf(real eps_local){

   interval tk_old, tk_new;
   real  dr_old, dr_new, factor;  
   int  k, k_index;


   if (ym < m_max && ydivide){     
     ri.f.taylor(0, ym + m_stride, interval(xinf, xsup), interval(yinf, ysup));
     tk_old  = ri.f.taylor(0, ym);
   }
   else
     tk_old = ri.f.taylor(0, ym, interval(xinf, xsup), interval(yinf, ysup));

    factor  = pow( (ysup - yinf) * 0.5, ym + 1 ) * (xsup - xinf) 
              * diam(tk_old);

   // looking for appropriate n
 
   dr_old = Sup(peano[ym_index][n_dim]) * factor; 
   if (ym != 2 * n_max) dr_old *= 2.0; 

   if (dr_old < eps_local){
    
     for(k_index = n_dim - 1, k = n_max - n_stride; 
         k_index >= peano_n[ym_index]; k_index--, k -= n_stride)
     {
        dr_new = Sup(peano[ym_index][k_index]) * factor;
        if (ym != 2 * k) dr_new *= 2.0;
        if(dr_new > eps_local) break;  
     }
     set_yriemann(tk_old, k_index + 1, k + n_stride);
     ydivide = false;           
     return;    
   }

   if (!ydivide) cout << " ydivideerror? " << endl;
   ydivide = true;
   // looking for appropriate m
   if (ym_index > 0){  
     k = ym - m_stride; k_index = ym_index -1;
     tk_new = power( (ysup - interval(yinf,yinf) ) * 0.5, k + 1) * 
              (xsup - interval(xinf, xinf)) * ri.f.taylor(0, k);
     if ( 2 * diam(tk_new * peano[k_index][n_dim]) < dr_old){ 
       set_yriemann(k_index, k, tk_new);          
       return; 
     }
   }

   if(ym < m_max){ 
     k = ym + m_stride; k_index = ym_index + 1;
     tk_new = power( ( ysup - interval(yinf, yinf) ) * 0.5, k + 1) * 
        (xsup - interval(xinf, xinf)) * ri.f.taylor(0, k); 
     if (k != 2 * n_max) dr_old *= 0.5; 
     if (diam(tk_new * peano[k_index][n_dim])  < dr_old){ 
       set_yriemann(k_index, k, tk_new);  
       return; 
    }
   }
   set_yriemann(tk_old, yn_index, yn);
}

// ----------------------------------------------------------------------------
// --- class cauchyxriemann_integral  -----------------------------------------
// ----------------------------------------------------------------------------


// --- cauchyxriemann_integral::constructor -----------------------------------

cauchyxriemann_integral::cauchyxriemann_integral(integrand& g, real xl): 
                                 f_c(cauchy::m_max, cauchy::m_max), 
                                 f_ctransposed(cauchy::m_max, cauchy::m_max){
 
 operand x, y(operand::Var_2);

 xlambda = xl;
 f_c = g;
 f   = f_c / ( x - xlambda);
 ftransposed = f.transpose();
 f_ctransposed = f_c.transpose();

 // cout << f << endl << ftransposed << endl;
 // cout << f_c << endl << f_ctransposed << endl;

 eps_rel = 0.0;
 if (!cauchy_counter)  xcauchy::init_data();
 cauchy_counter++;
}


cauchyxriemann_integral::cauchyxriemann_integral(): 
                                 f_c(cauchy::m_max, cauchy::m_max), 
                                 f_ctransposed(cauchy::m_max, cauchy::m_max){
 
 eps_rel = 0.0;
 if (!cauchy_counter)  xcauchy::init_data();
 cauchy_counter++;
}


// --- cauchyxriemann_integral::destructor ------------------------------------


cauchyxriemann_integral::~cauchyxriemann_integral(){ 
  cauchy_counter--;
  if (!cauchy_counter)  xcauchy::free_data();
}


// --- cauchyxriemann_integral::start_division --------------------------------

void cauchyxriemann_integral::start_division(){
  int  m_index  = (cauchy::m_max - cauchy::m_min) / (cauchy::m_stride),
       rm_index = (riemann::m_max - riemann::m_min) / (riemann::m_stride),
       n_index  = (riemann::n_max - riemann::n_min) / (riemann::n_stride);
 
  dotprecision r_dot;

  real p;

  r_dot = -xinf;
  accumulate(r_dot, -1.0, xsup);
  accumulate(r_dot,  2.0, xlambda); 
  if ( r_dot == 0.0)  //  one interval!
    new cauchyxriemann(xinf, xsup, yinf, ysup, m_index, n_index, m_index,
       n_index, *this, true, true, 0.5 * eps_abs);

  else if ( r_dot >= 0.0){ // lambda right
    r_dot = -xsup;
    accumulate(r_dot,  2.0, xlambda);
    rnd(r_dot, p);
    if (r_dot != p) throw xcauchy::no_screennumber(xinf, xsup, yinf, ysup); 

    // riemann interval
    new riemann(xinf, p, yinf, ysup, rm_index, n_index, rm_index, n_index,
                 *this, true, true, 0.5 * eps_abs * (p - xinf) / (xsup -xinf));

    // cauchy interval
    new cauchyxriemann(p, xsup, yinf, ysup, m_index, n_index, m_index, n_index,
                 *this, true, true, 0.5 * eps_abs * (xsup - p) / (xsup -xinf));
  }

  else{ // lambda left
    r_dot = -xinf;
    accumulate(r_dot,  2.0, xlambda);
    rnd(r_dot, p);
    if (r_dot != p) throw xcauchy::no_screennumber(xinf, xsup, yinf, ysup);

    new cauchyxriemann(xinf, p, yinf, ysup, m_index, n_index, m_index, n_index,
               *this, true, true, 0.5 * eps_abs * (p - xinf) / (xsup - xinf) );

    new riemann(p, xsup, yinf, ysup, rm_index, n_index, rm_index, n_index, 
                *this, true, true, 0.5 * eps_abs * (xsup - p) / (xsup - xinf));
  }
} 


// --- cauchyxriemann_integral::integrate -------------------------------------


interval cauchyxriemann_integral::integrate(interval a, interval b, interval c,
                                            interval d, real epsabs){

  xcauchy* tmpc;
  xriemann* tmpr;

  int  m_index  = xcauchy::m_dim, 
       rm_index = xriemann::m_dim,
       n_index  = xriemann::n_dim;

  cauchyxriemann_integral& ci = static_cast<cauchyxriemann_integral&>(*this);
  integral& ri = static_cast<integral&>(*this);

  eps_abs = epsabs;
  epsilon = 2 * eps_abs / ( (Sup(b) -Inf(a)) * (Sup(d) - Inf(c)) );

  xinf = Sup(a); xsup = Inf(b);
  yinf = Sup(c); ysup = Inf(d);
  Approximation = 0;
  // special intervals
  if (Inf(a) != Sup(a)){
    tmpr = new xriemann(Sup(c), Inf(d), Inf(a), Sup(a), rm_index, n_index,  
                              ri , epsilon, true);
    if (Inf(b) != Sup(b)){
        tmpr = new xriemann(Sup(c), Inf(d), Inf(b), Sup(b), rm_index, n_index, 
                                                        ri , epsilon, true); 
        if (Inf(c) != Sup(c)){
          tmpr = new xriemann(Inf(a),Sup(a),Inf(c),Sup(c), ri);
          tmpr = new xriemann(Inf(b),Sup(b),Inf(c),Sup(c), ri);
          tmpc = new xcauchy(Sup(a),Inf(b),Inf(c),Sup(c), 
                              m_index, n_index, ci, epsilon);
	}
        if (Inf(d) != Sup(d)){
          tmpr = new xriemann(Inf(a),Sup(a),Inf(d),Sup(d), ri);
          tmpr = new xriemann(Inf(b),Sup(b),Inf(d),Sup(d), ri);
          tmpc = new xcauchy(Sup(a),Inf(b),Inf(d),Sup(d), 
                              m_index, n_index, ci, epsilon);
	}      
    }
    else{
        if (Inf(c) != Sup(c)){
          tmpr = new xriemann(Inf(a),Sup(a),Inf(c),Sup(c), ri);
          tmpc = new xcauchy(Sup(a),Inf(b),Inf(c),Sup(c), 
                              m_index, n_index, ci, epsilon); 
	}
        if (Inf(d) != Sup(d)){
          tmpr = new xriemann(Inf(a),Sup(a),Inf(d),Sup(d),ri);
          tmpc = new xcauchy(Sup(a),Inf(b),Inf(d),Sup(d), 
                              m_index, n_index, ci, epsilon);
	}      
    }
  }
  else{
    if (Inf(b) != Sup(b)){
        tmpr = new xriemann(Sup(c), Inf(d), Inf(b), Sup(b), rm_index, n_index, 
                                                        ri , epsilon, true); 
        if (Inf(c) != Sup(c)){
          tmpr = new xriemann(Inf(b),Sup(b),Inf(c),Sup(c),ri);
          tmpc = new xcauchy(Sup(a),Inf(b),Inf(c),Sup(c), 
                                       m_index, n_index, ci, epsilon);
	}
        if (Inf(d) != Sup(d)){
          tmpr = new xriemann(Inf(b),Sup(b),Inf(d),Sup(d),ri);
          tmpc = new xcauchy(Sup(a),Inf(b),Inf(d),Sup(d), 
                                       m_index, n_index, ci, epsilon);
	}      
    }
    else{
        if (Inf(c) != Sup(c)) tmpc = new xcauchy(Sup(a), Inf(b), Inf(c), 
                                     Sup(c), m_index, n_index, ci, epsilon);
        if (Inf(d) != Sup(d)) tmpc = new xcauchy(Sup(a), Inf(b), Inf(d),
                                     Sup(d), m_index, n_index, ci, epsilon);
	}      
  }
  start_division();
  gaa();
 
  real tmp_inf, tmp_sup;
  interval result;

  rnd(Rinf, tmp_inf, RND_DOWN);
  rnd(Rsup, tmp_sup, RND_UP );
  rnd(Approximation, result);
  return result + interval(tmp_inf, tmp_sup);    

}


// ----------------------------------------------------------------------------
// --- class cauchy_integral  -------------------------------------------------
// ----------------------------------------------------------------------------


// --- cauchy_integral::constructor -------------------------------------------


cauchy_integral::cauchy_integral(integrand& g, real xl, real yl): 
  f_cc(cauchy::m_max, cauchy::m_max){
 
 operand x, y(operand::Var_2);

 xlambda = xl;
 ylambda = yl;

 f_cc = g;
 f   = f_cc / ( ( x - xlambda) * (y - ylambda) );
 ftransposed = f.transpose();
 f_c = f_cc / (y - ylambda);
 f_ctransposed = f_cc.transpose() / (y - xlambda);

 // cout << f << endl << ftransposed << endl;
 // cout << f_c << endl << f_ctransposed << endl;

 eps_rel = 0.0;
}


// --- cauchy_integral::start_division ----------------------------------------

void cauchy_integral::start_division(){

 int  cm_index = cauchy::m2m_index( cauchy::m_max),
      m_index  = riemann::m2m_index(riemann::m_max),
      n_index  = riemann::n_dim;
 
  real p, rxinf, rxsup, cxinf, cxsup, ryinf, rysup, cyinf, cysup, eps_local;

  bool xmid = false, ymid = false;
 
  dotprecision r_dot;

  // x division:
  r_dot = - xinf;
  accumulate(r_dot, -1.0, xsup);
  accumulate(r_dot,  2.0, xlambda);

  if ( r_dot == 0.0){  xmid = true;    // xlambda midpoint of x_interval 
   cxinf = xinf; cxsup = xsup;
  }

  else if ( r_dot >= 0.0){            //  xlambda right  
    r_dot = -xsup;
    accumulate(r_dot,  2.0, xlambda);
    rnd(r_dot, p);
    if (r_dot != p) throw cauchy::no_screennumber(xinf, xsup, yinf, ysup);
    cxinf = p;  cxsup = xsup; rxinf = xinf; rxsup = p;
  }

  else{                               // xlambda left 
    r_dot = -xinf;
    accumulate(r_dot,  2.0, xlambda);
    rnd(r_dot, p);
    if (r_dot != p) throw cauchy::no_screennumber(xinf, xsup, yinf, ysup);
    cxinf = xinf;  cxsup = p; rxinf = p; rxsup = xsup;
  }

  // y division: 
  r_dot = -yinf;
  accumulate(r_dot, -1.0, ysup);
  accumulate(r_dot,  2.0, ylambda);

  if ( r_dot == 0.0){  ymid  = true; // ylambda midpoint of y_interval 
    cyinf = yinf; cysup = ysup;
  }

  else if ( r_dot >= 0.0){ // ylambda right  
    r_dot = -ysup;
    accumulate(r_dot,  2.0, ylambda);
    rnd(r_dot, p);
    if (r_dot != p) throw cauchy::no_screennumber(xinf, xsup, yinf, ysup);
    cyinf = p;  cysup = ysup; ryinf = yinf; rysup = p;
  }

  else{  // ylambda left 
    r_dot = -yinf;
    accumulate(r_dot,  2.0, ylambda);
    rnd(r_dot, p);
    if (r_dot != p) throw cauchy::no_screennumber(xinf, xsup, yinf, ysup); 
    cyinf = yinf;  cysup = p; ryinf = p; rysup = ysup;
  }

  // generate cauchy subinterval 

  cauchy* cstart = new cauchy(cxinf, cxsup, cyinf, cysup, cm_index, n_index, 
                                                   cm_index, n_index,  *this);

  cstart->choose_cf(epsilon * (cxsup -  cxinf) * (cysup - cyinf) );
  push(*cstart);

  // generate subintervals 
  // special cases xmid or ymid
  if(xmid)
   if (ymid) return;
   else{
     new cauchyxriemann(xinf, xsup, ryinf, rysup, m_index, n_index, m_index, 
       n_index, *this, true, true, epsilon * (xsup - xinf) * (rysup - ryinf) );
     return; 
   }
  else if(ymid){
     new cauchyxriemann(yinf, ysup, rxinf, rxsup, m_index, n_index, m_index,
                        n_index, *this, true, true, 
                        epsilon * (rxsup -  rxinf) * (ysup - yinf), true);
     return; 
  } 
  
  
 
  new riemann(rxinf, rxsup, ryinf, rysup, m_index, n_index, m_index, n_index,
             *this, true, true, epsilon * (rxsup - rxinf) * (rysup - ryinf));
  new cauchyxriemann(cyinf, cysup, rxinf, rxsup, m_index, n_index, 
                     m_index, n_index, *this, true, true, 
                     epsilon * (rxsup - rxinf) * (cysup - cyinf), true);
  new cauchyxriemann(cxinf, cxsup, ryinf, rysup, m_index, n_index, m_index, 
                     n_index, *this, true, true, 
                     epsilon * (cxsup - cxinf) * (rysup - ryinf));
}

// --- cauchy_integral::integrate ---------------------------------------------

interval cauchy_integral::integrate(interval a, interval b, interval c,
                                            interval d, real epsabs){

  xcauchy* tmpc;
  xriemann* tmpr;

  int  m_index  = xcauchy::m_dim, 
    // rm_index = xriemann::m_dim,
       n_index  = xriemann::n_dim;

  cauchyxriemann_integral& ci = *this;
  //  static_cast<cauchyxriemann_integral&>(*this);
  integral& ri =  *this;
    // static_cast<integral&>(*this);
  eps_abs = epsabs;
  epsilon = 2 * eps_abs / ( (Sup(b) -Inf(a)) * (Sup(d) - Inf(c)) );
  Approximation = 0;

  xinf = Sup(a); xsup = Inf(b);
  yinf = Sup(c); ysup = Inf(d);
  Approximation = 0;
  // special intervals
  if (Inf(a) != Sup(a)){ 
    tmpc = new xcauchy(Sup(c), Inf(d), Inf(a), Sup(a), m_index, n_index,  
                              ci , epsilon, true);
    if (Inf(b) != Sup(b)){
        tmpc = new xcauchy(Sup(c), Inf(d), Inf(b), Sup(b), m_index, n_index, 
                                                        ci , epsilon, true); 
        if (Inf(c) != Sup(c)){
          tmpr = new xriemann(Inf(a),Sup(a),Inf(c),Sup(c), ri);
          tmpr = new xriemann(Inf(b),Sup(b),Inf(c),Sup(c), ri);
          tmpc = new xcauchy(Sup(a),Inf(b),Inf(c),Sup(c), 
                              m_index, n_index, ci, epsilon);
	}
        if (Inf(d) != Sup(d)){
          tmpr = new xriemann(Inf(a),Sup(a),Inf(d),Sup(d), ri);
          tmpr = new xriemann(Inf(b),Sup(b),Inf(d),Sup(d), ri);
          tmpc = new xcauchy(Sup(a),Inf(b),Inf(d),Sup(d), 
                              m_index, n_index, ci, epsilon);
	}      
    }
    else{
        if (Inf(c) != Sup(c)){
          tmpr = new xriemann(Inf(a),Sup(a),Inf(c),Sup(c), ri);
          tmpc = new xcauchy(Sup(a),Inf(b),Inf(c),Sup(c), 
                              m_index, n_index, ci, epsilon); 
	}
        if (Inf(d) != Sup(d)){
          tmpr = new xriemann(Inf(a),Sup(a),Inf(d),Sup(d),ri);
          tmpc = new xcauchy(Sup(a),Inf(b),Inf(d),Sup(d), 
                              m_index, n_index, ci, epsilon);
	}      
    }
  }
  else{
    if (Inf(b) != Sup(b)){
        tmpc = new xcauchy(Sup(c), Inf(d), Inf(b), Sup(b), m_index, n_index, 
                                                        ci , epsilon, true); 
        if (Inf(c) != Sup(c)){
          tmpr = new xriemann(Inf(b),Sup(b),Inf(c),Sup(c),ri);
          tmpc = new xcauchy(Sup(a),Inf(b),Inf(c),Sup(c), 
                                       m_index, n_index, ci, epsilon);
	}
        if (Inf(d) != Sup(d)){
          tmpr = new xriemann(Inf(b),Sup(b),Inf(d),Sup(d),ri);
          tmpc = new xcauchy(Sup(a),Inf(b),Inf(d),Sup(d), 
                                       m_index, n_index, ci, epsilon);
	}      
    }
    else{
        if (Inf(c) != Sup(c)) tmpc = new xcauchy(Sup(a), Inf(b), Inf(c), 
                                     Sup(c), m_index, n_index, ci, epsilon);
        if (Inf(d) != Sup(d)) tmpc = new xcauchy(Sup(a), Inf(b), Inf(d),
                                     Sup(d), m_index, n_index, ci, epsilon);
	}      
  }
  start_division();
  gaa();
 
  real tmp_inf, tmp_sup;
  interval result;

  rnd(Rinf, tmp_inf, RND_DOWN);
  rnd(Rsup, tmp_sup, RND_UP );
  rnd(Approximation, result);
  return result + interval(tmp_inf, tmp_sup);    

}

// ----------------------------------------------------------------------------
// --- class xcauchy ------ ---------------------------------------------------
// ----------------------------------------------------------------------------


// --- xcauchy::set_xcauchy ---------------------------------------------------

void xcauchy::set_xcauchy(int k_i, int k, interval tk){

 interval r;
  xm_index   = k_i;
  xm         = k; 
  r          = peano[xm_index][xn_index] * (ysup - interval(yinf,ysup)) * tk;
  if (xm != 2 * xn + 1) r -= r;
  r_inf      = Inf(r);
  r_sup      = Sup(r);
  dremainder = diam(r);
  return;
}


void xcauchy::set_xcauchy(interval tk, int k_i, int k){

  interval r;
  xn_index    = k_i;
  xn          = k; 
  r           = peano[xm_index][xn_index] * (ysup - interval(yinf,ysup)) *
                power((xsup- interval(xinf, xinf))* 0.5, xm) * tk;
  if (xm != 2 * xn + 1) r  -= r; 
  r_inf       = Inf(r);
  r_sup       = Sup(r);
  dremainder  = diam(r);
  return;
}



// --- xcauchy::add_approxsum -------------------------------------------------

void xcauchy::add_approxsum(){

   interval  *x   = riemann::node[xn_index] , 
             *w   = riemann::weight[xn_index],
             m, m2, c;
   int i,j;

   //    cauchy_integral& ci = dynamic_cast<cauchy_integral&>(ri);
   cauchy_integral& ci = static_cast<cauchy_integral&>(ri);
   integrand* fc = &ci.f_c;
   if (transposed) fc = &ci.f_ctransposed;

   m  = (xsup - interval(xinf, xinf)) * 0.5;
   c  = (xsup + interval(xinf, xinf)) * 0.5;
   m2 =  ysup - interval(yinf,ysup);

   if (xn % 2){
    j = xn / 2;

    accumulate(ci.Approximation, w[j] * m * m2, 
               fc->taylor(1, 0, m * x[j] + c, interval(xinf,xsup)));
    for(i=0; i<j; i++) 
      accumulate(ci.Approximation, w[i] / x[i] * m2 , 
              fc->value(m * x[i] + c, interval(xinf,xsup) ) ); 
    for(i=j+1; i<xn; i++)
      accumulate(ci.Approximation, w[i] / x[i] * m2, 
              fc->value(m * x[i] + c, interval(xinf,xsup) ) ); 
   }
   else for(i=0; i<xn ; i++)
     accumulate(ci.Approximation, w[i] / x[i] * m2 , 
              fc->value(m * x[i] + c, interval(xinf,xsup) ) ); 
}



// --- xcauchy::choose_qf -----------------------------------------------------

void xcauchy::choose_qf(const real& eps_local){

   //    cauchy_integral& ci = dynamic_cast<cauchy_integral&>(ri);
   cauchy_integral& ci = static_cast<cauchy_integral&>(ri); 
   integrand* fc = &ci.f_c;
   if (transposed) fc = &ci.f_ctransposed;

   interval tk_old, tk_new;
   real  dr_old, dr_new, factor;  
   int  k, k_index;

   if (xm < m_max){      // workaround:
     fc->taylor(xm + m_stride, interval(xinf, xsup), interval(yinf, ysup));
     tk_old  = fc->taylor(xm);
   } 
   else
     tk_old  = fc->taylor( xm, interval(xinf, xsup), interval(yinf, ysup) ); 
   factor  = power( (xsup - xinf) * 0.5, xm ) * diam(tk_old);

   // looking for appropriate n
   
   dr_old = Sup(peano[xm_index][riemann::n_dim]) * factor; 
   if (xm != 2 * xn + 1) dr_old *= 2.0; 

   if (dr_old < eps_local){
    
     for(k_index = riemann::n_dim - 1, k = riemann::n_max - riemann::n_stride; 
         k_index >= 0; k_index--, k -= riemann::n_stride)
     {
        dr_new = Sup(peano[xm_index][k_index]) * factor;
        if (xm != 2 * k) dr_new *= 2.0;
        if(dr_new > eps_local) break;  
     }
     set_xcauchy(tk_old, k_index + 1, k + riemann::n_stride);          
     return;    
   }

   // looking for appropriate m
   if (xm_index > 0){  
     k = xm - m_stride; k_index = xm_index -1;
     tk_new = power( (xsup - interval(xinf,xinf) ) * 0.5, k ) 
                                                * fc->taylor(k);
     if ( 2 * diam(tk_new * peano[k_index][riemann::n_dim]) < dr_old){ 
       set_xcauchy(k_index, k, tk_new);          
       return; 
     }
   }

   if(xm < m_max){ 
     k = xm + m_stride; k_index = xm_index + 1;
     tk_new = power(( xsup - interval(xinf, xinf) ) * 0.5, k ) * fc->taylor(k);
    if (k != 2 * riemann::n_max + 1) dr_old *= 0.5; 
    if (diam(tk_new * peano[k_index][riemann::n_dim])  < dr_old){ 
       set_xcauchy(k_index, k, tk_new);  
       return; 
    }
   }
   set_xcauchy(tk_old, xn_index, xn);
}



// --- xcauchy::subdivide -----------------------------------------------------

void xcauchy::subdivide(){
 // trisection: three equal parts
 //    cauchy_integral& ci = dynamic_cast<cauchy_integral&>(ri);

   cauchy_integral& ci = static_cast<cauchy_integral&>(ri); 

   real eps_local = ci.eps_abs * (xsup -xinf) / ( 2.0 * (ci.xsup - ci.xinf))
                    / (ci.ysup - ci.yinf),
    //  eps_local = ci.eps_abs * (xsup -xinf) / ( 4.0 * (ci.xsup - ci.xinf)),
        p1, p2;
   xsubdivide(p1, p2);
  
   // left part riemann interval    
   xriemann* left = new xriemann(xinf, p1, yinf, ysup, xm_index, 
                                 xn_index, ci, transposed);
   left->choose_qf(eps_local);
   ci.push(*left);

   // right part riemann interval
   xriemann* right = new xriemann(p2, xsup, yinf, ysup,xm_index, 
                                  xn_index, ci, transposed);
   right->choose_qf(eps_local);
   ci.push(*right);

   // cauchy interval
   xinf = p1;
   xsup = p2; 
   choose_qf( 2 * eps_local);
   approx = false;
   ci.push(*this);
}

 
// --- xcauchy::init_data -----------------------------------------------------


void xcauchy::init_data(){

// init peanoconstants
  int i, j;
  idotprecision idot;

  peano   = new interval*[m_dim + 1];
  peano_n = new int[m_dim + 1];
  peano_w = new interval[riemann::n_dim + 1];

  for(i = 0; i <= m_dim; i++) 
    peano[i] = new interval[riemann::n_dim + 1];
 
  read_data("./data/peanocauchy8.hex" , peano, peano_n, m_max, m_min, m_stride,
                                        riemann::n_min, riemann::n_stride);

  read_data("./data/peanocauchy11.hex", peano, peano_n, m_max, m_min, m_stride,
                                        riemann::n_min, riemann::n_stride);

  read_data("./data/peanocauchy14.hex", peano, peano_n, m_max, m_min, m_stride,
                                        riemann::n_min, riemann::n_stride);

  read_data("./data/peanocauchy17.hex", peano, peano_n, m_max, m_min, m_stride,
                                        riemann::n_min, riemann::n_stride);

  for(int j=0; j<=m_dim; j++){
    i = -1;
    do{ 
      if (i< riemann::n_dim) i++;
      else throw ;                 //  remove!!!!!         
    }
    while(peano[j][i] == interval(0,0));
    peano_n[j] = i;
  }

  for(i=0; i<= riemann::n_dim; i++){
   idot = 0;
   for(j = 0; j < riemann::n_index2n(i) / 2; j++)
      accumulate(idot, riemann::weight[i][j] / riemann::node[i][j], -1.0);
   rnd(idot, peano_w[i]);
  }
  
}


// --- xcauchy::free_data() ---------------------------------------------------


void xcauchy::free_data(){

  for(int i = 0; i <= m_dim; i++) 
    delete[] peano[i];
  delete[] peano;
  delete[] peano_n;
  delete[] peano_w;
} 

// --- xcauchy::xsubdivide ----------------------------------------------------

void xcauchy::xsubdivide(real& left, real& right){

   r_dot = 0;
   if (xsup >= - xinf){    // xlambda > 0
     accumulate(r_dot, 0.25,  xinf);
     accumulate(r_dot, 0.75,  xsup);
     rnd(r_dot, right);
     r_dot = 0.0; 
     accumulate(r_dot,  1.0,  xsup);
     accumulate(r_dot,  1.0,  xinf);
     accumulate(r_dot, -1.0,  right);
     rnd(r_dot, left); 
     if (r_dot != left) 
        throw no_screennumber(xinf, xsup, yinf, ysup);  
   }
   else{
     accumulate(r_dot, 0.75,  xinf);
     accumulate(r_dot, 0.25,  xsup);
     rnd(r_dot, left);
     r_dot = 0; 
     accumulate(r_dot,  1.0, xsup);
     accumulate(r_dot,  1.0, xinf);
     accumulate(r_dot, -1.0, left);
     rnd(r_dot, right); 
     if (r_dot != right) 
        throw no_screennumber(xinf, xsup, yinf, ysup);  
   }
}




// ----------------------------------------------------------------------------
// --- class cauchyxriemann ---------------------------------------------------
// ----------------------------------------------------------------------------





// --- cauchyxriemann::subdivide() --------------------------------------------

void cauchyxriemann::subdivide(){

 cauchyxriemann_integral& ci = static_cast<cauchyxriemann_integral&>(ri);
 cauchyxriemann* c_interval;
 riemann* r;
 real eps_local =  ci.epsilon 
                   * (xsup - xinf) * (ysup -yinf),
      left = xinf, right = xsup, ymid;
  
 if (xdivide){  
   xsubdivide(left, right);
   if (ydivide){      // 6 new intervals
      r_dot = 0;
      accumulate(r_dot, 0.5,  yinf);
      accumulate(r_dot, 0.5,  ysup);
      rnd(r_dot, ymid);
      eps_local /= 6.0;
      if (transposed){
        new riemann(yinf, ymid, xinf, left, ym_index, yn_index, ym_index, 
                    yn_index, ri, ydivide, true, eps_local);
        new riemann(ymid, ysup, xinf, left, ym_index, yn_index, ym_index, 
                    yn_index, ri, ydivide, true, eps_local);   
        new riemann(yinf, ymid, right, xsup, ym_index, yn_index, ym_index, 
                    yn_index, ri, ydivide, true, eps_local);
        new riemann(ymid, ysup, right, xsup, ym_index, yn_index, ym_index,
                    yn_index, ri, ydivide, true, eps_local);
      }
      else{
        new riemann(xinf, left, yinf, ymid, ym_index, yn_index, ym_index, 
                    yn_index, ri, true, ydivide, eps_local);
        new riemann(xinf, left, ymid, ysup, ym_index, yn_index, ym_index, 
                    yn_index, ri, true, ydivide, eps_local);   
        new riemann(right, xsup, yinf, ymid, ym_index, yn_index, ym_index, 
                    yn_index, ri, true, ydivide, eps_local);
        new riemann(right, xsup, ymid, ysup, ym_index, yn_index, ym_index,
                    yn_index, ri, true, ydivide, eps_local);
      }
      new cauchyxriemann(left, right, yinf, ymid, xm_index, xn_index, ym_index,
                  yn_index, ci, xdivide, ydivide, eps_local, transposed);

      xinf = left; xsup = right;  yinf = ymid;   
      choose_xcf(eps_local);
      //      choose_ycf(eps_local);
       if (xn % 2) choose_ycfodd(eps_local);
       else choose_ycfeven(eps_local);
      ci.push(*this);   
   } 
   else{              // 3 new intervals

      eps_local /= 3.0;
      if(transposed){
        new riemann(yinf, ysup, xinf, left, ym_index, yn_index, ym_index, 
                    yn_index, ri, true, true, eps_local);
        new riemann(yinf, ysup, right, xsup, ym_index, yn_index, ym_index, 
                    yn_index, ri, true, true, eps_local);

      }
      else{
        new riemann(xinf, left, yinf, ysup, ym_index, yn_index, ym_index, 
                    yn_index, ri, true, true, eps_local);
        new riemann(right, xsup, yinf, ysup, ym_index, yn_index, ym_index, 
                    yn_index, ri, true, true, eps_local);
      }
      xinf = left; xsup = right;
      choose_xcf(eps_local);
      //      choose_ycf(eps_local);
      if (xn % 2) choose_ycfodd(eps_local);
      else choose_ycfeven(eps_local);
      ci.push(*this);   
   }
 }
 else if(ydivide){  // 2 new intervals

      r_dot = 0;
      accumulate(r_dot, 0.5,  yinf);
      accumulate(r_dot, 0.5,  ysup);
      rnd(r_dot, ymid);
      eps_local *= 0.5;
  
      new cauchyxriemann(xinf, xsup, yinf, ymid, xm_index, xn_index, 
             ym_index, yn_index, ci, xdivide, ydivide, eps_local, transposed);

      yinf = ymid;
      choose_xcf(eps_local);
      //     choose_ycf(eps_local);
       if (xn % 2) choose_ycfodd(eps_local);
       else choose_ycfeven(eps_local);
      ci.push(*this);  
     }
 else{ ci.push(*this);  xdivide = ydivide = true;}     
}



// --- cauchyxriemann::set_xcaxri ---------------------------------------------


inline void cauchyxriemann::set_xcaxri(int k_i, int k, interval tk){
 
  interval r;
  xm_index    = k_i;
  xm          = k; 
  r           = cauchy::peano[xm_index][xn_index] * tk;
  if (xm != 2 * xn + 1) r -= r;
  r_inf      = Inf(r);
  r_sup      = Sup(r);
  dremainder = diam(r);
  return;

}



inline void cauchyxriemann::set_xcaxri(interval tk, int k_i, int k){

 interval r;
  xn_index    = k_i;
  xn          = k; 
  r          = cauchy::peano[xm_index][xn_index] * 
               power((xsup- interval(xinf, xinf))* 0.5, xm) * 
               (ysup - interval(yinf, yinf)) * tk;
  if (xm != 2 * xn + 1) r  -= r; 
  r_inf      = Inf(r);
  r_sup      = Sup(r);
  dremainder = diam(r);
  return;
}


// --- cauchyxriemann::set_ycaxri ---------------------------------------------


inline void cauchyxriemann::set_ycaxri(interval tk, interval tk1, int k_i, 
                                                                  int k){
 interval r1, r2;
  yn_index   = k_i;
  yn         = k; 

  r2         = power((ysup- interval(yinf, yinf))* 0.5, ym + 1); 
  r1         = r2 * tk * peano_w[xn_index];
  r2        *= (xsup - interval(xinf, xinf))* 0.5 * 
               weight[xn_index][xn_index /2] * tk1; 

 
  if (ym != 2 * yn){ r1   = 2 * r1 + r2; r1 -= r1;}
  else               r1  += r2 - r1;
  
  r1         *= riemann::peano[ym_index][yn_index];
  r_inf      += Inf(r1);
  r_sup      += Sup(r1);
  dremainder += diam(r1);
  return;
}  


inline void cauchyxriemann::set_ycaxri(int k_i, int k, interval tk, 
                                                       interval tk1){
  interval r;
   ym_index   = k_i;
   ym         = k; 

   if (ym != 2 * yn){ r = 2 * tk + tk1; r -= r; }
   else r = tk - tk + tk1;

   r *= riemann::peano[ym_index][yn_index];
   
   r_inf      += Inf(r);
   r_sup      += Sup(r);
   dremainder += diam(r);
   return;
}



inline void cauchyxriemann::set_ycaxri(interval tk, int k_i, int k){
 interval r1;
  yn_index   = k_i;
  yn         = k; 

  r1         = (xsup - interval(xinf, xinf)) * 
               power( (ysup- interval(yinf, yinf))* 0.5, ym + 1) * tk;
 
  if (ym != 2 * yn) r1  -= r1;
  
  r1         *= riemann::peano[ym_index][yn_index];
  r_inf      += Inf(r1);
  r_sup      += Sup(r1);
  dremainder += diam(r1);
  return;
}  


inline void cauchyxriemann::set_ycaxri(int k_i, int k, interval tk){
  interval r;
   ym_index   = k_i;
   ym         = k; 
   
   
   if (ym != 2 * yn) tk -= tk;

   tk *= riemann::peano[ym_index][yn_index];
   
   r_inf      += Inf(tk);
   r_sup      += Sup(tk);
   dremainder += diam(tk);
   return;
}


// --- cauchyxriemann::choose_xcf() -------------------------------------------

void cauchyxriemann::choose_xcf(const real& eps_local){

  // choose xm and xn
   interval tk_old, tk_new;
   real     dr_old, dr_new, factor;  
   int      k, k_index;

   cauchyxriemann_integral& ci = static_cast<cauchyxriemann_integral&>(ri); 
   integrand* fc      = &ci.f_c;
   if (transposed) fc = &ci.f_ctransposed;

   if (xm < cauchy::m_max && xdivide){      
     fc->taylor(xm + cauchy::m_stride, interval(xinf, xsup), 
               interval(yinf, ysup));
     tk_old  = fc->taylor(xm);
   }
   else 
     tk_old = fc->taylor(xm, interval(xinf, xsup), interval(yinf, ysup));

   factor = power( (xsup - xinf) * 0.5, xm) * (ysup - yinf) * diam(tk_old);

   // looking for appropriate n
 
   dr_old = Sup(cauchy::peano[xm_index][n_dim]) * factor;
 
   if (xm != 2 * n_max + 1) dr_old *= 2.0; 

   if (dr_old < eps_local){
    
     for(k_index = n_dim - 1, k = n_max - n_stride; 
         k_index >= cauchy::peano_n[xm_index]; k_index--, k -= n_stride)
     {
        dr_new = Sup(cauchy::peano[xm_index][k_index]) * factor;
        if (xm != 2 * k + 1) dr_new *= 2.0;
        if(dr_new > eps_local) break;  
     }
     set_xcaxri(tk_old, k_index + 1, k + n_stride);
     xdivide = false;          
     return;    
   }
   if (!xdivide) cout << "xdivide error ? " << endl;
   xdivide = true;
   // looking for appropriate m
   if (xm_index > 0){  
     k = xm - cauchy::m_stride; k_index = xm_index -1;
     tk_new = power( (xsup - interval(xinf,xinf) ) * 0.5, k) * 
              (ysup - interval(yinf, yinf)) * fc->taylor(k);
     if ( 2 * diam(tk_new * cauchy::peano[k_index][n_dim]) <  dr_old){ 
       set_xcaxri(k_index, k, tk_new);          
       return; 
     }
   }

   if(xm < cauchy::m_max){ 
     k = xm + cauchy::m_stride; k_index = xm_index + 1;
     tk_new = power( ( xsup - interval(xinf, xinf) ) * 0.5, k) * 
       (ysup - interval(yinf, yinf)) * fc->taylor(k); 
     if (k != 2 * n_max + 1) dr_old *= 0.5; 
     if (diam(tk_new * cauchy::peano[k_index][n_dim])  < dr_old){ 
       set_xcaxri(k_index, k, tk_new);  
       return; 
    }
   }
   set_xcaxri(tk_old, xn_index, xn);
}

 
// --- cauchyxriemann::choose_ycf() -------------------------------------------


void cauchyxriemann::choose_ycf(const real& eps_local){
  // choose ym and yn for even xn

   interval tk_old, tk_new;
   real     dr_old, dr_new, factor;  
   int      k, k_index;

   cauchyxriemann_integral& ci = static_cast<cauchyxriemann_integral&>(ri); 
   integrand* fc = &ci.f_c;
   if (transposed) fc = &ci.f_ctransposed;

   if (ym < riemann::m_max && ydivide){
      fc->taylor(1, ym + riemann::m_stride, interval(xinf, xsup), 
                     interval(yinf, ysup));
      tk_old  = fc->taylor(1, ym);
   }
   else 
     tk_old = fc->taylor(1, ym,interval(xinf, xsup), interval(yinf, ysup));

   factor = power( (ysup - yinf) * 0.5, ym + 1) * (xsup - xinf)
              * diam(tk_old);

   // looking for appropriate n
 
   dr_old = Sup(riemann::peano[ym_index][n_dim]) * factor;
 
   if (ym != 2 * n_max) dr_old *= 2.0; 

   if (dr_old < eps_local){
    
     for(k_index = n_dim - 1, k = n_max - n_stride; 
         k_index >= riemann::peano_n[ym_index]; k_index--, k -= n_stride)
     {
        dr_new = Sup(riemann::peano[ym_index][k_index]) * factor;
        if (ym != 2 * k) dr_new *= 2.0;
        if(dr_new > eps_local) break;  
     }
     set_ycaxri(tk_old, k_index + 1, k + n_stride);
     ydivide = false;          
     return;    
   }

   if (!ydivide) cout << "ydivide error ? " << endl;
   ydivide = true;

   // looking for appropriate m

   if (ym_index > 0){  
     k = ym - riemann::m_stride; k_index = ym_index - 1;

     tk_new = power( (ysup - interval(yinf,yinf) ) * 0.5, k + 1) * 
              (xsup - interval(xinf, xinf)) *  fc->taylor(1, k);

     if ( diam(tk_new * riemann::peano[k_index][n_dim]) <  dr_old){ 
       set_ycaxri(k_index, k, tk_new);          
       return; 
     }
   }

   if(ym < riemann::m_max){ 
     k = ym + riemann::m_stride; k_index = ym_index + 1;

     tk_new = power( ( ysup - interval(yinf, yinf) ) * 0.5, k + 1) * 
              (xsup - interval(xinf, xinf)) * fc->taylor(1, k); 

     if (k != 2 * n_max) dr_old *= 0.5; 

     if (diam(tk_new * riemann::peano[k_index][n_dim]) < dr_old){ 
       set_ycaxri(k_index, k, tk_new);  
       return; 
    }
   }
   set_ycaxri(tk_old, yn_index, yn);
   return;
}



// --- cauchyxriemann::choose_ycfodd() ----------------------------------------


void cauchyxriemann::choose_ycfodd(const real& eps_local){
  // choose ym and yn for odd xn

   interval tk_old, tk_old1, tk_new, tk_new1;
   real     dr_old, dr_new, factor, factor2;  
   int      k, k_index;

   cauchyxriemann_integral& ci = static_cast<cauchyxriemann_integral&>(ri); 
   integrand* fc = &ci.f_c;
   if (transposed) fc = &ci.f_ctransposed;

   if (ym < riemann::m_max && ydivide){
      fc->taylor(1, ym + riemann::m_stride, interval(xinf, xsup), 
                        interval(yinf, ysup));
      tk_old1  = fc->taylor(1, ym);
   }
   else 
     tk_old1 = fc->taylor(1, ym,interval(xinf, xsup), interval(yinf, ysup));

   tk_old     = fc->taylor(0, ym);

   // estimate the diam of the remainder 
   factor2   =  power( (ysup - yinf) * 0.5, ym + 1);
   factor    =  factor2 * diam(peano_w[xn_index] * tk_old);
   factor2  *=  (xsup - xinf) * 0.5 * 
                diam(weight[xn_index][xn /2] * tk_old1);
 
   factor    =  2.0 * factor + factor2; 

   dr_old = Sup(riemann::peano[ym_index][n_dim]) * factor;   
   if (ym != 2 * n_max) dr_old *= 2.0; 

   if (dr_old < eps_local){
     for(k_index = n_dim - 1, k = n_max - n_stride; 
         k_index >= riemann::peano_n[ym_index]; k_index--, k -= n_stride)
     {
       dr_new = Sup(riemann::peano[ym_index][k_index]) * factor;
        if (ym != 2 * k) dr_new *= 2.0;
        if(dr_new > eps_local) break;  
     }
     set_ycaxri(tk_old, tk_old1, k_index + 1, k + n_stride);
     ydivide = false;          
     return;    
   }
   
   if (!ydivide) cout << "ydivide error ? " << endl;
   ydivide = true;
   // looking for appropriate m
   if (ym_index > 0){  

     k = ym - riemann::m_stride; k_index = ym_index - 1;

     tk_new1  = power( (ysup - interval(yinf,yinf) ) * 0.5, k + 1);

     tk_new   = tk_new1  * fc->taylor(0, k) * peano_w[xn_index];
     tk_new1 *= (xsup - interval(xinf,xinf)) * 0.5 * fc->taylor(1, k)
                * weight[xn_index][xn /2];  
     dr_new   = 2 * Sup(riemann::peano[k_index][n_dim]) * 
                ( 2 *  diam(tk_new) + diam(tk_new1));                

     if ( dr_new  <  dr_old){ 
       set_ycaxri(k_index, k, tk_new, tk_new1);          
       return; 
     }
   }
   if(ym < riemann::m_max){
 
     k = ym + riemann::m_stride; k_index = ym_index + 1;

     tk_new1  = power( (ysup - interval(yinf,yinf) ) * 0.5, k + 1);

     tk_new   = tk_new1 * fc->taylor(0, k) * peano_w[xn_index];
     tk_new1 *= (xsup - interval(xinf,xinf)) * 0.5 *  fc->taylor(1, k)
                * weight[xn_index][xn /2]; 

     dr_new   =  Sup(riemann::peano[k_index][n_dim]) * 
                 (2 *  diam(tk_new) + diam(tk_new1));
     if (k != 2 * n_max) dr_new *= 2; 
     if (dr_new < dr_old){ 
       set_ycaxri(k_index, k, tk_new, tk_new1);  
       return; 
    }
   }
   set_ycaxri(tk_old, tk_old1, yn_index, yn);
   return;
} 



// --- cauchyxriemann::choose_ycfeven() ---------------------------------------


void cauchyxriemann::choose_ycfeven(const real& eps_local){
  // choose ym and yn for even xn

   interval tk_old, tk_new;
   real     dr_old, dr_new, factor;  
   int      k, k_index;

   cauchyxriemann_integral& ci = static_cast<cauchyxriemann_integral&>(ri); 
   integrand* fc = &ci.f_c;
   if (transposed) fc = &ci.f_ctransposed;

   if (ym < riemann::m_max && ydivide){
      fc->taylor(0, ym + riemann::m_stride, interval(xinf, xsup), 
                     interval(yinf, ysup));
      tk_old  = fc->taylor(0, ym);
   }
   else 
     tk_old = fc->taylor(0, ym,interval(xinf, xsup), interval(yinf, ysup));

   factor = power( (ysup - yinf) * 0.5, ym + 1) * 2.0 * 
                diam( peano_w[xn_index] * tk_old );

   // looking for appropriate n
 
   dr_old = Sup(riemann::peano[ym_index][n_dim]) * factor;
 
   if (ym != 2 * n_max) dr_old *= 2.0; 

   if (dr_old < eps_local){
    
     for(k_index = n_dim - 1, k = n_max - n_stride; 
         k_index >= riemann::peano_n[ym_index]; k_index--, k -= n_stride)
     {
        dr_new = Sup(riemann::peano[ym_index][k_index]) * factor;
        if (ym != 2 * k) dr_new *= 2.0;
        if(dr_new > eps_local) break;  
     }
     set_ycaxri(tk_old, interval(0,0), k_index + 1, k + n_stride);
     ydivide = false;          
     return;    
   }

   if (!ydivide) cout << "ydivide error ? " << endl;
   ydivide = true;

   // looking for appropriate m

   if (ym_index > 0){  
     k = ym - riemann::m_stride; k_index = ym_index - 1;

     tk_new = power( (ysup - interval(yinf,yinf) ) * 0.5, k + 1) * 
              peano_w[xn_index] * fc->taylor(0, k);

     if ( 2 * diam(tk_new * riemann::peano[k_index][n_dim]) <  dr_old){ 
       set_ycaxri(k_index, k, tk_new, interval(0,0));          
       return; 
     }
   }

   if(ym < riemann::m_max){ 
     k = ym + riemann::m_stride; k_index = ym_index + 1;

     tk_new = power( ( ysup - interval(yinf, yinf) ) * 0.5, k + 1) * 
              peano_w[xn_index] * fc->taylor(0, k); 

     if (k != 2 * n_max) dr_old *= 0.5; 

     if (diam(tk_new * riemann::peano[k_index][n_dim]) < dr_old){ 
       set_ycaxri(k_index, k, tk_new, interval(0,0));  
       return; 
    }
   }
   set_ycaxri(tk_old, interval(0,0), yn_index, yn);
   return;
}



// --- cauchyxriemann::add_approxsum ------------------------------------------


void cauchyxriemann::add_approxsum(){ 

  cauchyxriemann_integral& ci = static_cast<cauchyxriemann_integral&>(ri);

  integrand* fc = &ci.f_c;
  if (transposed) fc = &ci.f_ctransposed;

  int i,j, i2 = xn / 2;

  interval  *x   = riemann::node[xn_index] ,  *y  = riemann::node[yn_index], 
            *xw  = riemann::weight[xn_index], *yw = riemann::weight[yn_index],
            xm, xc, ym, yc;
interval help;
 help = ci.Approximation;
  xm  = (xsup - interval(xinf, xinf)) * 0.5;
  xc  = (xsup + interval(xinf, xinf)) * 0.5;
  ym  = (ysup - interval(yinf, yinf)) * 0.5;
  yc  = (ysup + interval(yinf, yinf)) * 0.5;
  
  if(xn%2){


  for(i=0; i < i2 ; i++)
      for(j = 0; j < yn ; j++)
        accumulate(ci.Approximation, xw[i] / x[i] * yw[j] * ym, 
                   fc->value(xm * x[i] + xc, ym * y[j] + yc));
  for(i= (xn + 1) / 2; i < xn; i++)
      for(j = 0; j < yn ; j++)
        accumulate(ci.Approximation, xw[i] / x[i] * yw[j] * ym, 
                   fc->value(xm * x[i] + xc, ym * y[j] + yc));
  for(j = 0; j < yn ; j++)
       accumulate(ci.Approximation, xw[i2] * xm  * yw[j] * ym, 
                   fc->taylor(1, 0, xm * x[i2] + xc, ym * y[j] + yc));
  }
  else{
      for(i=0; i < xn ; i++)
         for(j = 0; j < yn ; j++)
             accumulate(ci.Approximation, xw[i] / x[i] * yw[j] *ym, 
                   fc->value(xm * x[i] + xc, ym * y[j] + yc));
  }
help = ci.Approximation;
}



// ----------------------------------------------------------------------------
// --- class cauchy -----------------------------------------------------------
// ----------------------------------------------------------------------------



// --- cauchy::add_approxsum --------------------------------------------------


void cauchy::add_approxsum(){ 

  cauchy_integral& ci = static_cast<cauchy_integral&>(ri);

  int i,j, i2 = xn / 2, j2 = yn / 2;

  interval  *x   = riemann::node[xn_index] ,  *y  = riemann::node[yn_index], 
            *xw  = riemann::weight[xn_index], *yw = riemann::weight[yn_index],
            xm, xc, ym, yc;
  
  xm  = (xsup - interval(xinf, xinf)) * 0.5;
  xc  = (xsup + interval(xinf, xinf)) * 0.5;
  ym  = (ysup - interval(yinf, yinf)) * 0.5;
  yc  = (ysup + interval(yinf, yinf)) * 0.5;



  for(i=0; i < i2 ; i++){

    for(j = 0; j < j2 ; j++)
      accumulate(ci.Approximation, xw[i] / x[i] * yw[j] / y[j], 
                 ci.f_cc.value(xm * x[i] + xc, ym * y[j] + yc));
    for(j = (yn + 1) / 2; j < yn ; j++)
      accumulate(ci.Approximation, xw[i] / x[i] * yw[j] / y[j], 
                 ci.f_cc.value(xm * x[i] + xc, ym * y[j] + yc)); 
 }

 for(i= (xn + 1) / 2; i < xn; i++){

   for(j = 0; j < j2 ; j++)
     accumulate(ci.Approximation, xw[i] / x[i] * yw[j] / y[j], 
                ci.f_cc.value(xm * x[i] + xc, ym * y[j] + yc));
   for(j = (yn + 1) / 2; j < yn ; j++)
     accumulate(ci.Approximation, xw[i] / x[i] * yw[j] / y[j], 
                ci.f_cc.value(xm * x[i] + xc, ym * y[j] + yc));
 }
 
 if (xn % 2){    // xn odd
    for(j = 0; j < j2 ; j++)
      accumulate(ci.Approximation, xw[i2] * xm * yw[j] / y[j], 
                 ci.f_cc.taylor(1, 0, xm * x[i2] + xc, ym * y[j] + yc));
    for(j = (yn + 1) / 2; j < yn ; j++)
      accumulate(ci.Approximation, xw[i2] * xm * yw[j] / y[j], 
                 ci.f_cc.taylor(1, 0, xm * x[i2] + xc, ym * y[j] + yc)); 
 
    if (yn % 2){ // xn odd yn odd
      accumulate(ci.Approximation, xw[i2] * xm * ym * yw[j2], 
                 ci.f_cc.taylor(1, 1, xm * x[i2] + xc, ym * y[j2] + yc));  
      for(i=0; i < i2 ; i++)
         accumulate(ci.Approximation, xw[i] / x[i] * yw[j2] * ym, 
                ci.f_cc.taylor(0, 1, xm * x[i] + xc, ym * y[j2] + yc));
      for(i= (xn + 1) / 2; i < xn; i++)
         accumulate(ci.Approximation, xw[i] / x[i] * yw[j2] * ym, 
                ci.f_cc.taylor(0, 1, xm * x[i] + xc, ym * y[j2] + yc));
    }
    return;
  }
  if (yn % 2){  // xn even yn odd
    for(i=0; i < xn; i++)
         accumulate(ci.Approximation, xw[i] / x[i] * yw[j2] * ym, 
                ci.f_cc.taylor(0, 1, xm * x[i] + xc, ym * y[j2] + yc));
   }
 interval help;
  help = ci.Approximation;
}





// --- cauchy::subdivide() ----------------------------------------------------

void cauchy::subdivide(){

 cauchy_integral& ci = static_cast<cauchy_integral&>(ri);

  real cxinf, cxsup, cyinf, cysup, 
       Xinf = xinf, Xsup = xsup, 
       Yinf = yinf, Ysup = Ysup,
       eps_local = ci.epsilon;

 

  xsubdivide(cxinf,cxsup);
  ysubdivide(cyinf,cysup);


  if (choose_cf(eps_local * (xsup - xinf) * (ysup -yinf) )){ // xdivide
    new cauchyxriemann(cxinf, cxsup, yinf, cyinf, xm_index, xn_index, 
                       ym_index, yn_index, ci, true, true, 
                       eps_local * (cxsup - cxinf) * (cyinf - yinf) );
    new cauchyxriemann(cxinf, cxsup, cysup, ysup, xm_index, xn_index,
                       ym_index, yn_index, ci, true, true,
                       eps_local * (cxsup - cxinf) * (ysup - cysup) );
    new cauchyxriemann(yinf, ysup, xinf, cxinf, xm_index, xn_index, 
                       ym_index, yn_index, ci, true, true,
                       eps_local * (cxinf - xinf) * (ysup - yinf), true);
    new cauchyxriemann(yinf, ysup, cxsup, xsup, xm_index, xn_index, 
                       ym_index, yn_index, ci, true, true,
                       eps_local * (xsup - cxsup) * (ysup - yinf), true);
  }
  else{ // ydivide

    new cauchyxriemann( xinf, xsup, yinf, cyinf, xm_index, xn_index, 
                        ym_index, yn_index, ci, true, true,
                        eps_local * (xsup - xinf) * (cyinf - yinf) ); 
    new cauchyxriemann( xinf, xsup, cysup, ysup,  xm_index, xn_index, 
                        ym_index, yn_index, ci, true, true,
                        eps_local * (xsup - xinf) * (ysup - cysup) );
    new cauchyxriemann( cyinf, cysup, xinf, cxinf, xm_index, xn_index, 
                        ym_index, yn_index, ci, true, true,
                        eps_local * (cxinf - xinf) * (cysup - cyinf), true);
    new cauchyxriemann( cyinf, cysup, cxsup,  xsup, xm_index, xn_index, 
                        ym_index, yn_index, ci, true, true,
                        eps_local * (xsup - cxsup) * (cysup - cyinf), true);
  }

  xinf = cxinf; xsup = cxsup;
  yinf = cyinf; ysup = cysup;
  ci.push(*this);
}
 


// --- cauchy::remainders -----------------------------------------------------


void cauchy::remainders(interval& Ex1, interval& Ex2, interval& Ey1,
                                       interval& Ey2, interval& Exy){

  xn      =  xriemann::n_index2n(xn_index),
  yn      =  xriemann::n_index2n(yn_index);
     
  Exy  = peano[xm_index][xn_index] * peano[ym_index][yn_index] * Exy;
  Ex1  = peano_w[yn_index] * peano[xm_index][xn_index] * (Ex1 - Ex1); 
  if( yn % 2) 
     Ex2 = riemann::weight[yn_index][yn / 2] * peano[xm_index][xn_index] * Ex2;
  else 
     Ex2 = 0.0;

  if( xm == 2 * xn + 1) Ex1 += Ex2 ; 
  else{                 Ex1 += Ex1 + Ex2 - Ex2;  Exy -= Exy;} 

  Ey1 = peano_w[xn_index] * peano[ym_index][yn_index] * (Ey1 - Ey1);

  if( xn % 2) 
     Ey2 = riemann::weight[xn_index][xn / 2] * peano[ym_index][yn_index] * Ey2;
  else 
     Ey2 = 0.0;

  if (ym == 2 * yn + 1) Ey1 += Ey2;
  else{                 Ey1 += Ey1 + Ey2 -Ey2; Exy -= Exy;}
}



// --- cauchy::set_tk ---------------------------------------------------------

void cauchy::set_tk(interval& xtk, interval& xtk1, interval& ytk,
                                   interval& ytk1, interval& xytk){

  cauchy_integral& ci = static_cast<cauchy_integral&>(ri);

  xm = m_index2m(xm_index);
  ym = m_index2m(ym_index);

  interval xh = power( (xsup - interval(xinf, xinf) ) * 0.5, xm),
           yh = power( (ysup - interval(yinf, yinf) ) * 0.5, ym);

   xtk   = ci.f_cc.taylor(xm,  0, interval(xinf, xsup), interval(yinf, ysup))
           * xh;
   ytk   = ci.f_cc.taylor(0,  ym, interval(xinf, xsup), interval(yinf, ysup))
           * yh;
   xytk  = ci.f_cc.taylor(xm, ym, interval(xinf, xsup), interval(yinf, ysup))
           * xh * yh; 
   xtk1  = ci.f_cc.taylor(xm,  1, interval(xinf, xsup), interval(yinf, ysup))
           * xh * ( ysup - interval(yinf, yinf) ) * 0.5;
   ytk1  = ci.f_cc.taylor(1,  ym, interval(xinf, xsup), interval(yinf, ysup))
           * yh * ( xsup - interval(xinf, xinf) ) * 0.5;
}
 

 
// --- cauchy::choose_cf() ----------------------------------------------------
// -- Rxy(f) = RxQy + RyQx + RxRy ---------------------------------------------
/*

bool cauchy::choose_cf(const real& eps_local){

  interval xtk, ytk, xytk, Ex, Ey, Exy, r,
           xtk1, ytk1;

  int k1, k2, k3, k4;

  bool xdivide;

  cauchy_integral& ci = static_cast<cauchy_integral&>(ri);

  if (xm < m_max) k1 = xm + m_stride;
  else k1 = xm;
  if (ym < m_max) k2 = ym + m_stride;
  else k2 = ym;

  ci.f_cc.taylor(k1, k2, interval(xinf, xsup), interval(yinf, ysup));
  set_tk(xtk, xtk1, ytk, ytk1, xytk);

  Ex  = xtk; 
  Ey  = ytk; 
  Exy = xytk;

  k3 = xn_index = peano_n[xm_index];
  k4 = yn_index = peano_n[ym_index];

  remainders(Ex, xtk1, Ey, ytk1, Exy);

  r = Ex + Ey + Exy;

  if ( diam(r) < eps_local ){
     r_inf = Inf(r); r_sup = Sup(r); dremainder = diam(r);
     return( Sup(Ex) > Sup(Ey) );
  }

  yn_index = peano_n[ym_index] + 1;

  for(; xn_index <= riemann::n_dim; xn_index++){ 

     if (xn_index != peano_n[xm_index]) yn_index = peano_n[ym_index];

     for(; yn_index <= riemann::n_dim; yn_index++){ 
        Ex = xtk; Ey = ytk; Exy = xytk;
        remainders(Ex, xtk1, Ey, ytk1, Exy);
        Exy += Ex + Ey;
        if ( diam(Exy) < eps_local ){
           r_inf = Inf(Exy); r_sup = Sup(Exy); dremainder = diam(Exy);
           return ( Sup(Ex) > Sup(Ey) );
	}
        if ( diam(Exy) < diam(r) ){
           r = Exy;
           k3 = xn_index; k4 = yn_index;
	}
     }  
  }

  xn_index = k3;
  yn_index = k4;
  k1       = m2m_index(k1);
  k2       = m2m_index(k2);
  xm_index = 0; 
  ym_index = 0; 
 
  set_tk(xtk, xtk1, ytk, ytk1, xytk);
  Ex = xtk; Ey = ytk; Exy = xytk;
  remainders(Ex, xtk1, Ey, ytk1, Exy);

  r = Exy + Ex + Ey;

  if ( diam(r) < eps_local ){
     r_inf = Inf(r); r_sup = Sup(r); dremainder = diam(r);
     return ( Sup(Ex) < Sup(Ey) );
  }

  xdivide = ( Ex > Ey );
  ym_index = 1; 

  for( ; xm_index <= k1; xm_index++){
    if (xm_index) ym_index = 0;
    for(; ym_index <= k2; ym_index++){
       set_tk(xtk, xtk1, ytk, ytk1, xytk);
       Ex = xtk; Ey = ytk; Exy = xytk;
       remainders(Ex, xtk1, Ey, ytk1, Exy);
       Exy += Ex + Ey;
       if ( diam(Exy) < eps_local ){       
         r_inf = Inf(Exy); r_sup = Sup(Exy); dremainder = diam(Exy);
         return (Sup(Ex) > Sup(Ey) );
       }
       if ( diam(Exy) < diam(r) ){
         r = Exy;
         k3 = xm_index; k4 = ym_index;
         xdivide = ( Sup(Ex) > Sup(Ey) );
       }
    }
  }

  xm_index = k3; xm = m_index2m(k3); xn = riemann:: n_index2n(xn_index);
  ym_index = k4; ym = m_index2m(k4); yn = riemann:: n_index2n(yn_index);
  r_inf = Inf(r); r_sup = Sup(r); dremainder = diam(r);
  return xdivide;
}
*/


// --- cauchy::set_xcauchy ----------------------------------------------------


inline real cauchy::set_xcauchy(int k_i, int k, interval tk){
 
  interval r;
  xm_index    = k_i;
  xm          = k; 
  r           = cauchy::peano[xm_index][xn_index] * tk;
  if (xm != 2 * xn + 1) r -= r;
  r_inf      = Inf(r);
  r_sup      = Sup(r);
  dremainder = diam(r);
  return diam(r);

}



inline real cauchy::set_xcauchy(interval tk, int k_i, int k){

 interval r;
  xn_index    = k_i;
  xn          = k; 
  r          = cauchy::peano[xm_index][xn_index] * 
               power((xsup- interval(xinf, xinf))* 0.5, xm) * 
               (ysup - interval(yinf, yinf)) * tk;
  if (xm != 2 * xn + 1) r  -= r; 
  r_inf      = Inf(r);
  r_sup      = Sup(r);
  dremainder = diam(r);
  return diam(r);
}



// --- cauchy::set_ycauchy ----------------------------------------------------


inline real cauchy::set_ycauchy(interval tk, interval tk1, int k_i, int k){
 interval r1, r2;
  yn_index   = k_i;
  yn         = k; 

  r2         = power((ysup- interval(yinf, yinf))* 0.5, ym); 
  r1         = r2 * tk * peano_w[xn_index];
  r2        *= (xsup - interval(xinf, xinf))* 0.5 * 
               riemann::weight[xn_index][xn_index /2] * tk1; 

 
  if (ym != 2 * yn + 1){ r1   = 2 * r1 + r2; r1 -= r1;}
  else               r1  += r2 - r1;
  
  r1         *= cauchy::peano[ym_index][yn_index];
  r_inf      += Inf(r1);
  r_sup      += Sup(r1);
  dremainder += diam(r1);
  return diam(r1); 
}  


inline real cauchy::set_ycauchy(int k_i, int k, interval tk, interval tk1){
  interval r;
   ym_index   = k_i;
   ym         = k; 

   if (ym != 2 * yn + 1){ r = 2 * tk + tk1; r -= r; }
   else r = tk - tk + tk1;

   r *= cauchy::peano[ym_index][yn_index];
   
   r_inf      += Inf(r);
   r_sup      += Sup(r);
   dremainder += diam(r);
   return diam(r);
}


// --- cauchy::choose_xcf() ---------------------------------------------------

real cauchy::choose_xcf(const real& eps_local){

  // choose xm and xn
   interval tk_old, tk_new;
   real     dr_old, dr_new, factor;  
   int      k, k_index, n_dim = xriemann::n_dim,
            n_max = xriemann::n_max, n_stride = xriemann::n_stride;

   cauchy_integral& ci = static_cast<cauchy_integral&>(ri); 
   integrand* f      = &ci.f_cc;

   if (xm < cauchy::m_max){      
     f->taylor(xm + cauchy::m_stride, 1, interval(xinf, xsup), 
               interval(yinf, ysup));
     tk_old  = f->taylor(xm, 1);
   }
   else 
     tk_old = f->taylor(xm, 1, interval(xinf, xsup), interval(yinf, ysup));

   factor = power( (xsup - xinf) * 0.5, xm) * (ysup - yinf) * diam(tk_old);

   // looking for appropriate n
 
   dr_old = Sup(cauchy::peano[xm_index][xriemann::n_dim]) * factor;
 
   if (xm != 2 * xriemann::n_max + 1) dr_old *= 2.0; 

   if (dr_old < eps_local){
    
     for(k_index = xriemann::n_dim - 1, k = xriemann::n_max -  n_stride; 
         k_index >= cauchy::peano_n[xm_index]; k_index--, k -= n_stride)
     {
        dr_new = Sup(cauchy::peano[xm_index][k_index]) * factor;
        if (xm != 2 * k + 1) dr_new *= 2.0;
        if(dr_new > eps_local) break;  
     }
     return set_xcauchy(tk_old, k_index + 1, k + n_stride);
   }

   // looking for appropriate m
   if (xm_index > 0){  
     k = xm - cauchy::m_stride; k_index = xm_index -1;
     tk_new = power( (xsup - interval(xinf,xinf) ) * 0.5, k) * 
              (ysup - interval(yinf, yinf)) * f->taylor(k,1);
     if ( 2 * diam(tk_new * cauchy::peano[k_index][n_dim]) <  dr_old){ 
       return set_xcauchy(k_index, k, tk_new);          
     }
   }

   if(xm < cauchy::m_max){ 
     k = xm + cauchy::m_stride; k_index = xm_index + 1;
     tk_new = power( ( xsup - interval(xinf, xinf) ) * 0.5, k) * 
       (ysup - interval(yinf, yinf)) * f->taylor(k,1); 
     if (k != 2 * n_max + 1) dr_old *= 0.5; 
     if (diam(tk_new * cauchy::peano[k_index][n_dim])  < dr_old){ 
       return set_xcauchy(k_index, k, tk_new);  
    }
   }
   return set_xcauchy(tk_old, xn_index, xn);
}



// --- cauchy::choose_ycfodd() ------------------------------------------------
// ----------------------------------------------------------------------------
// x_n odd  enclosure of RyQx(f)

real cauchy::choose_ycfodd(const real& eps_local){
  // choose ym and yn for odd xn

   interval tk_old, tk_old1, tk_new, tk_new1;
   real     dr_old, dr_new, factor, factor2;  
   int      k, k_index, n_dim = xriemann::n_dim,
            n_max = xriemann::n_max, n_stride = xriemann::n_stride;

   cauchy_integral& ci = static_cast<cauchy_integral&>(ri); 
   integrand* f = &ci.f_cc;

   if (ym < cauchy::m_max){
      f->taylor(1, ym + cauchy::m_stride, interval(xinf, xsup), 
                        interval(yinf, ysup));
      tk_old1  = f->taylor(1, ym);
   }
   else 
     tk_old1 = f->taylor(1, ym,interval(xinf, xsup), interval(yinf, ysup));

   tk_old     = f->taylor(0, ym);

   // estimate the diam of the remainder 
   factor2   =  power( (ysup - yinf) * 0.5, ym);
   factor    =  factor2 * diam(peano_w[xn_index] * tk_old);
   factor2  *=  (xsup - xinf) * 0.5 * 
                diam(xriemann::weight[xn_index][xn /2] * tk_old1);
 
   factor    =  2.0 * factor + factor2; 

   dr_old = Sup(cauchy::peano[ym_index][n_dim]) * factor;   
   if (ym != 2 * n_max + 1) dr_old *= 2.0; 

   if (dr_old < eps_local){
     for(k_index = n_dim - 1, k = n_max - n_stride; 
         k_index >= cauchy::peano_n[ym_index]; k_index--, k -= n_stride)
     {
       dr_new = Sup(cauchy::peano[ym_index][k_index]) * factor;
        if (ym != 2 * k) dr_new *= 2.0;
        if(dr_new > eps_local) break;  
     }
     return set_ycauchy(tk_old, tk_old1, k_index + 1, k + n_stride);           
   }
   
   // looking for appropriate m
   if (ym_index > 0){  

     k = ym - cauchy::m_stride; k_index = ym_index - 1;

     tk_new1  = power( (ysup - interval(yinf,yinf) ) * 0.5, k);

     tk_new   = tk_new1  * f->taylor(0, k) * peano_w[xn_index];
     tk_new1 *= (xsup - interval(xinf,xinf)) * 0.5 * f->taylor(1, k)
                * xriemann::weight[xn_index][xn /2];  
     dr_new   = 2 * Sup(cauchy::peano[k_index][n_dim]) * 
                ( 2 *  diam(tk_new) + diam(tk_new1));                

     if ( dr_new  <  dr_old){ 
       return set_ycauchy(k_index, k, tk_new, tk_new1);          
     }
   }
   if(ym < cauchy::m_max){
 
     k = ym + cauchy::m_stride; k_index = ym_index + 1;

     tk_new1  = power( (ysup - interval(yinf,yinf) ) * 0.5, k);

     tk_new   = tk_new1 * f->taylor(0, k) * peano_w[xn_index];
     tk_new1 *= (xsup - interval(xinf,xinf)) * 0.5 *  f->taylor(1, k)
                * xriemann::weight[xn_index][xn /2]; 

     dr_new   =  Sup(cauchy::peano[k_index][n_dim]) * 
                 (2 *  diam(tk_new) + diam(tk_new1));
     if (k != 2 * n_max + 1) dr_new *= 2; 
     if (dr_new < dr_old){ 
       return set_ycauchy(k_index, k, tk_new, tk_new1);  
    }
   }
   return set_ycauchy(tk_old, tk_old1, yn_index, yn);
} 



// --- cauchy::choose_ycfeven() -----------------------------------------------
// ----------------------------------------------------------------------------
// x_n even  enclosure of RyQx(f)

real cauchy::choose_ycfeven(const real& eps_local){
// choose ym and yn for even xn

   interval tk_old, tk_new;
   real     dr_old, dr_new, factor;  
   int      k, k_index, n_dim = xriemann::n_dim,
            n_max = xriemann::n_max, n_stride = xriemann::n_stride;

   cauchy_integral& ci = static_cast<cauchy_integral&>(ri); 
   integrand* f = &ci.f_cc;

   if (ym < cauchy::m_max){
      f->taylor(0, ym + cauchy::m_stride, interval(xinf, xsup), 
                     interval(yinf, ysup));
      tk_old  = f->taylor(0, ym);
   }
   else 
     tk_old = f->taylor(0, ym,interval(xinf, xsup), interval(yinf, ysup));

   factor = power( (ysup - yinf) * 0.5, ym) * 2.0 * 
                diam( peano_w[xn_index] * tk_old );

   // looking for appropriate n
 
   dr_old = Sup(cauchy::peano[ym_index][n_dim]) * factor;
 
   if (ym != 2 * n_max + 1) dr_old *= 2.0; 

   if (dr_old < eps_local){
    
     for(k_index = n_dim - 1, k = n_max - n_stride; 
         k_index >= cauchy::peano_n[ym_index]; k_index--, k -= n_stride)
     {
        dr_new = Sup(cauchy::peano[ym_index][k_index]) * factor;
        if (ym != 2 * k) dr_new *= 2.0;
        if(dr_new > eps_local) break;  
     }
     return set_ycauchy(tk_old, interval(0,0), k_index + 1, k + n_stride);    
   }

   // looking for appropriate m

   if (ym_index > 0){  
     k = ym - cauchy::m_stride; k_index = ym_index - 1;

     tk_new = power( (ysup - interval(yinf,yinf) ) * 0.5, k) * 
              peano_w[xn_index] * f->taylor(0, k);

     if ( 4 * diam(tk_new * cauchy::peano[k_index][n_dim]) <  dr_old){ 
       return set_ycauchy(k_index, k, tk_new);           
     }
   }

   if(ym < cauchy::m_max){ 
     k = ym + cauchy::m_stride; k_index = ym_index + 1;

     tk_new = power( ( ysup - interval(yinf, yinf) ) * 0.5, k) * 
              peano_w[xn_index] * f->taylor(0, k);
 
     dr_new = 2 * diam(tk_new * cauchy::peano[k_index][n_dim]);
     if (k != 2 * n_max) dr_new *= 2; 

     if (dr_new < dr_old){ 
       return set_ycauchy(k_index, k, tk_new);  
    }
   }
   return set_ycauchy(tk_old, interval(0,0), yn_index, yn);
}


// --- cauchy::choose_cf() ----------------------------------------------------
// -- Rxy(f) = RxIy + RyQx  ---------------------------------------------------
// the better one wins ...

bool cauchy::choose_cf(const real& eps_local){
 real tmp;
 tmp = choose_xcf(eps_local);
 if (xn % 2) return (tmp > choose_ycfodd(eps_local));
 else return (tmp > choose_ycfeven(eps_local));
}


// --- cauchy::ysubdivide -----------------------------------------------------

void cauchy::ysubdivide(real& left, real& right){

   r_dot = 0;
   if (ysup >= - yinf){    // ylambda > 0
     accumulate(r_dot, 0.25,  yinf);
     accumulate(r_dot, 0.75,  ysup);
     rnd(r_dot, right);
     r_dot = 0.0; 
     accumulate(r_dot,  1.0,  ysup);
     accumulate(r_dot,  1.0,  yinf);
     accumulate(r_dot, -1.0,  right);
     rnd(r_dot, left); 
     if (r_dot != left) 
        throw no_screennumber(xinf, xsup, yinf, ysup);  
   }
   else{
     accumulate(r_dot, 0.75,  yinf);
     accumulate(r_dot, 0.25,  ysup);
     rnd(r_dot, left);
     r_dot = 0; 
     accumulate(r_dot,  1.0, ysup);
     accumulate(r_dot,  1.0, yinf);
     accumulate(r_dot, -1.0, left);
     rnd(r_dot, right); 
     if (r_dot != right) 
        throw no_screennumber(xinf, xsup, yinf, ysup);  
   }
}

operand x, y(operand::Var_2);
