//============================================================================
//
//                              Program/Module
//                                   from
//                C++ package for verified numerical integration
//                  of singular and strong singular functions
//                      Copyright (c) 2000 Stefan Wedner
//
// This program/module is free software for non-commercial use.
//
//
//============================================================================
//----------------------------------------------------------------------------
//
// File:     quadrature.C
// See also: quadrature.h
// Purpose:  computing enclosures of one-dimensional integrals
//     
// ===========================================================================


#include<fstream>
#include"rmath.hpp"
#include"quadrature.h"


#ifdef  TIME_CHECK
#include<time.h> 
  clock_t  start_taylor, end_taylor, start_function, end_function;
#endif 

// ----------------------------------------------------------------------------
// --- constants and static elements-------------------------------------------
// ----------------------------------------------------------------------------

dotprecision base_interval::r_dot;


const int integral::max_heapsize = 5000;

const int riemann::n_min    =   8;
const int riemann::n_max    =  17;
const int riemann::n_stride =   3;
const int riemann::n_dim    = (riemann::n_max - riemann::n_min) 
                              / riemann::n_stride;
const int riemann::m_min    =   7;
const int riemann::m_max    =  16;
const int riemann::m_stride =   3;
const int riemann::m_dim    = (riemann::m_max - riemann::m_min) 
                              / riemann::m_stride;

interval ** riemann::node;
interval ** riemann::weight;
interval ** riemann::peano = riemann::init_data();

interval ** sqrt_interval::node;
interval ** sqrt_interval::weight;
interval ** sqrt_interval::peano;
int sqrt_integral::sqrt_counter   =  0;

const int sqrt_interval::n_min    =  6;
const int sqrt_interval::n_max    = 12;
const int sqrt_interval::n_stride =  3;
const int sqrt_interval::n_dim    =  2;
// remember values in start_division

const int cauchy::m_min        =  riemann::m_min + 1;
const int cauchy::m_max        =  riemann::m_max + 1;
const int cauchy::m_stride     =  riemann::m_stride;
const int cauchy::m_dim        =  riemann::m_dim;

interval ** cauchy::peano = cauchy::init_data();

const int hadamard::had_max = 7;    

interval* hadamard::peano = hadamard::init_data();

// ----------------------------------------------------------------------------
// --- auxiliary functions ----------------------------------------------------
// ----------------------------------------------------------------------------

void read_datasymm(char* name, interval** node, interval** weight, 
                   int n_min = 1, int n_stride = 1){

  int  i, n;
  real a_down, a_up;
  interval *pointer1, *pointer2;
  ifstream external_data(name);
  if(! external_data) cout << " error! : file " << name 
                           << " not available! " << endl;
  external_data >> n;

  pointer1   = node[(n - n_min) / n_stride];
  pointer2   = weight[(n - n_min) / n_stride];

  for(i=0; i<= n / 2 - 1; i++){
    external_data >> Hex >>  a_down;
    external_data >> Hex >>  a_up;
    pointer1[i] = interval(a_down,a_up);
    pointer1[n - i - 1] = -  pointer1[i]; 
  }
  if(n%2 ){
    external_data >> Hex >> a_down;
    external_data >> Hex >> a_up;
    pointer1[n/2] = interval(a_down,a_up);
  }
  for(i=0; i<= n / 2 - 1; i++){
    external_data >> Hex >>  a_down;
    external_data >> Hex >>  a_up;
    pointer2[i] = interval(a_down,a_up);
    pointer2[n - i - 1] =   pointer2[i]; 
  }
  if(n%2 ){
    external_data >> Hex >> a_down;
    external_data >> Hex >> a_up;
    pointer2[n/2] = interval(a_down,a_up);
  }  
  cout << Dec;
  external_data.close();
} 



void read_data(char* name, interval** node, interval** weight, 
               int n_min = 1, int n_stride = 1){
  int  i, n;
  real a_down, a_up;
  interval* pointer1, *pointer2;
  ifstream external_data(name);

  if(! external_data) cout << " error! : file " << name 
                           << " not available! " << endl;
  external_data >> n;

  pointer1   = node[(n - n_min) / n_stride];
  pointer2   = weight[(n - n_min) / n_stride];

  for(i=0; i< n; i++){
    external_data >> Hex >>  a_down;
    external_data >> Hex >>  a_up;
    pointer1[i] = interval(a_down,a_up);
  }
  for(i=0; i< n; i++){
    external_data >> Hex >>  a_down;
    external_data >> Hex >>  a_up;
    pointer2[i] = interval(a_down,a_up);
  }
  cout << Dec;
  external_data.close();
} 



int read_data(char* name, interval** peano, int m_max, int m_min = 1, 
              int m_stride = 1, int n_min = 1, int n_stride = 1){

  int  i, n, data_min, data_max;
  real a_down, a_up;
  ifstream external_data(name);
  if(! external_data) cout << " error! : file " << name 
                           << " not available! " << endl;
  
  external_data >> n;
  if (n < n_min) throw;

  external_data >> data_min;
  external_data >> data_max;

  if (data_max < m_max) throw;

  for(i= data_min; i< m_min; i++){
    external_data >> Hex >>  a_down;
    external_data >> Hex >>  a_up;
    external_data >> Hex >>  a_down;
    external_data >> Hex >>  a_up;
  }

  external_data >> Hex >>  a_down;
  external_data >> Hex >>  a_up;
  external_data >> Hex >>  a_down;
  external_data >> Hex >>  a_up;
  n = (n - n_min) / n_stride;
  peano[0][n]  = interval(a_down,a_up);
  
  for(i= 1; i<= (m_max - m_min) / m_stride; i++){
    for( int j=0; j < m_stride; j++){
      external_data >> Hex >>  a_down;
      external_data >> Hex >>  a_up;
      external_data >> Hex >>  a_down;
      external_data >> Hex >>  a_up;
    }
    peano[i][n]  = interval(a_down,a_up);
  }
  external_data.close();
  cout << Dec;
  return 0;
}





 
// ----------------------------------------------------------------------------
// --- class integral ---------------------------------------------------------
// ----------------------------------------------------------------------------

integral::integral() :f(riemann::m_max){

  heapsize = 0;
  heap = new pbase_interval[max_heapsize];
  fcounter = 0;

  Rinf = 0;
  Rsup = 0;
  Quadsum = 0;
  eps_rel = 0;
  eps_abs = 0;
}


integral::integral(integrand& g) :f(riemann::m_max){

  heapsize = 0;
  heap = new pbase_interval[max_heapsize];
  fcounter = 0;

  Rinf = 0;
  Rsup = 0;
  Quadsum = 0;
  f = g;
  eps_rel = 0;
  eps_abs = 0;
}


integral::~integral(){ 
  heapsize = 0;
  delete[] heap ;
}


void integral::start_division(){

  int m_index = (riemann::n_max - riemann::n_min) / riemann::n_stride,
      n_index = (riemann::n_max - riemann::n_min) / riemann::n_stride;

  riemann* start = new riemann(inf, sup, m_index, n_index, *this);
  start->choose_qf(eps_abs);
  push( *start);
}



interval integral::integrate(real a, real b, real epsabs){

  inf = a;
  sup = b;
  eps_abs = epsabs;
  if (fcounter)  throw; // user error1
  Quadsum = 0;
  start_division();
  gaa();
 
  real tmp_inf, tmp_sup;
  interval result;

  rnd(Rinf, tmp_inf, RND_DOWN);
  rnd(Rsup, tmp_sup, RND_UP);
  rnd(Quadsum, result);
  return result + interval(tmp_inf, tmp_sup);    
}


interval integral::integrate(real epsabs){
  fcounter = 0;
  Quadsum  = 0;
  eps_abs  = epsabs;
  gaa();

  real tmp_inf, tmp_sup;
  interval result;

  rnd(Rinf, tmp_inf, RND_DOWN);
  rnd(Rsup, tmp_sup, RND_UP);
  rnd(Quadsum, result);
  return result + interval(tmp_inf, tmp_sup);  
}



interval integral::integrate(interval a, interval b, real epsabs){

  riemann* tmp; 
  interval r;

  inf = Sup(a);
  sup = Inf(b);
  if (inf >= sup) throw bound_error();

  if (inf  != Inf(a)){
    // main value  
    tmp = new riemann(Inf(a), inf, 0, 0, *this); 
    tmp->quadsum = interval(0, Sup(a - a) ) * f.value(a);
    tmp->quad = true;
    tmp->n    = 1;
    // r =  tmp->quadsum;
    push( *tmp);
  }

  if (sup  != Sup(b)){
    // main value
    tmp = new riemann(sup, Sup(b), 0, 0, *this);
    tmp->quadsum = interval(0, Sup(b - b) ) * f.value(b);
    tmp->quad = true;
    tmp->n   = 1; 
    //  r += tmp->quadsum;
    push( *tmp);
  }

  interval result = integrate(inf, sup, epsabs);

  return result;    
}



interval integral::integrate(interval a, real b, real epsabs){
  if (Sup(a) >= b) throw bound_error();
  return integrate(a, interval(b,b), epsabs);
}



interval integral::integrate(real a, interval b, real epsabs){
  if (a >= Inf(b)) throw bound_error();
  return integrate(interval(a, a), b, epsabs);
}



void  integral::gaa(){

  base_interval* old_range;

#ifdef TIME_CHECK
  start_taylor = clock();
#endif 

  while( (Rsup - Rinf) > eps_abs && heapsize < max_heapsize ){
    old_range = pop();
    old_range->subdivide();
  }

#ifdef TIME_CHECK
  end_taylor = clock();
  start_function = clock();
#endif 
  // compute quadraturesum   
  for(int i=0; i<heapsize; i++){
     old_range = heap[i].pointer;
     if (old_range->quad) accumulate(Quadsum, 1.0, old_range->quadsum);
     else old_range->add_quadraturesum();
     fcounter += old_range->n;  
  }  

#ifdef TIME_CHECK
  end_function = clock();
  cout <<  " timetable " << endl;
  // cout <<  start_taylor << " , " << end_taylor << endl;
  cout << " t_R   " << end_taylor - start_taylor << endl  
       << " t_Q   " << end_function - start_function << endl 
       << " total " << end_function - start_taylor << endl;
  // double r = end_function - start_function ; 
 //  cout   <<  (end_taylor - start_taylor) /  r;
#endif 

  if (heapsize == max_heapsize){ cout << " HEAP" << endl; throw ;}
            // max heap_size reached! 
}


ostream& operator<<(ostream& os, const integral& out){
  interval Q, R;
  real rinf, rsup;
  rnd(out.Quadsum, Q); rnd(out.Rinf, rinf, RND_DOWN); 
  rnd(out.Rsup, rsup, RND_UP);
  R =  interval(rinf, rsup);
  os << Scientific; //" function: " << out.f << endl;
  // os << " integrationbounds:   " << out.inf << " , " << out.sup << endl;
  os << " number of intervals : " << out.heapsize << endl;
  os << " #f                  : " << out.fcounter << endl; 
  os << " approximationsum    : " << Q << endl
     << " d(approximationsum) : " << diam(Q) << endl;
  os << " remainder           : " << R << endl
     << " d(remainder)        : " << diam(R) << endl;
  Q += R;
  os << " enclosure           : " << Q << endl 
     << " d(enclosure)        : " << diam(Q) << endl;
}

// Tex Output:
// ostream& operator<<(ostream& os, const integral& out){
//    interval Q, R;
//    real rinf, rsup;
//    rnd(out.Quadsum, Q); rnd(out.Rinf, rinf, -1); rnd(out.Rsup, rsup, 1);
//    R =  interval(rinf, rsup);
//    Q += R;
//    os << " &  " << Inf(Q) << " , " 
//          << "\\hspace{-1ex}\\begin{array}{c}   \\"
//        << "\\[-0.75ex]" << "  \\end{array}" << "\\hspace{-1ex}E " 
//       << Sup(Q) 
//       << " & " << SetPrecision(5,3) << diam(Q) << " & " 
//       <<  out.heapsize <<  " & " << out.fcounter; 
// }


// show all details:

void integral::print_division(){

  base_interval* old_range;

  for(int i=0; i<heapsize; i++){
     old_range = heap[i].pointer;
     old_range->print_interval(cout);
  }
}



void integral::remainders(real a, real b){

  interval tk, factor, rest;

  int  m, m_index, n, n_index, m_min = riemann::m_min, 
       m_dim =  riemann::m_dim,  m_stride =  riemann:: m_stride,
    m_max = riemann::m_max,
        n_dim =  riemann::n_dim, n_min =  riemann::n_min,
        n_stride =  riemann:: n_stride  ;

  tk  = f.taylor(m_max, interval(a, b) ); 
  
  for(m_index = 0, m = m_min; 
      m_index <= m_dim; m_index++, m += m_stride){
    //  cout << a << " " << b << " " << m << endl << tk << endl;
  tk =  f.taylor(m);  
  factor  = power( (b-interval(a) ) * 0.5, m + 1 ) * tk ;
  cout << Scientific << SetPrecision(10,4) <<  " m " << m << endl;   
  cout << " tk " << tk << " peano " << riemann::peano[m_index][0] << endl;
  for(n_index = 0, n = n_min; 
      n_index <= n_dim; n_index++, n += n_stride)
    { // berechne restglied
       rest = riemann::peano[m_index][n_index] * factor;
       cout << n << endl << " remainder: " << rest 
            << " d: " << diam(rest) << endl; 
     }     
  }
}





// ----------------------------------------------------------------------------
// --- class base_interval ----------------------------------------------------
// ----------------------------------------------------------------------------


ostream& operator<<(ostream& os, const base_interval& out){
   os << " [" << out.inf << " , " << out.sup << " ]" << endl;
   os << " R " << out.r_inf  << " ,"  << out.r_sup 
      << " diam: " << out.dremainder;
   os << " m: " << out.m << " , n: " << out.n << endl;
   return os;
}



// ----------------------------------------------------------------------------
// --- class pbase_interval----------------------------------------------------
// ----------------------------------------------------------------------------
// pointer layer for base_interval (speed up heapsort!)
// 

ostream& operator<<(ostream& os, const pbase_interval& out){
  os << out.pointer->dremainder;
  return os;
}

istream& operator>>(istream& is, pbase_interval& in){
  is >> in.pointer->dremainder;
  return is;
}

bool  pbase_interval::operator<(const pbase_interval& t) const{
  return pointer->dremainder < t.pointer->dremainder;
}

bool  pbase_interval::operator>(const pbase_interval& t) const{
  return pointer->dremainder > t.pointer->dremainder;
}

bool  pbase_interval::operator==(const pbase_interval& t) const{
  return pointer->dremainder == t.pointer->dremainder;
}



// ----------------------------------------------------------------------------
// --- class riemann- ---------------------------------------------------------
// ----------------------------------------------------------------------------


interval** riemann::init_data(){

  // init node, weight, peano

  node   = new interval*[n_dim + 1];
  weight = new interval*[n_dim + 1];
  peano  = new interval*[m_dim + 1];

  for(int i = 0; i<= n_dim; i++){
    node[i]   = new interval[n_min + i * n_stride];
    weight[i] = new interval[n_min + i * n_stride];
  } 

  read_datasymm("./data/gaussleg8.hex" , node, weight, n_min, n_stride);
  read_datasymm("./data/gaussleg11.hex", node, weight, n_min, n_stride);
  read_datasymm("./data/gaussleg14.hex", node, weight, n_min, n_stride);
  read_datasymm("./data/gaussleg17.hex", node, weight, n_min, n_stride);
 
  for(int i = 0; i <= m_dim; i++) 
    peano[i] = new interval[n_dim + 1];
 
  read_data("./data/peanoleg8.hex", peano ,m_max , m_min, m_stride,
                                                   n_min, n_stride);

  read_data("./data/peanoleg11.hex", peano, m_max, m_min, m_stride,
                                                   n_min, n_stride);

  read_data("./data/peanoleg14.hex", peano, m_max, m_min, m_stride,
                                                   n_min, n_stride);

  read_data("./data/peanoleg17.hex", peano, m_max, m_min, m_stride,
                                                   n_min, n_stride);
  return peano;
}



void riemann::free_data(){
 int i;

  for(i = 0; i<= n_dim; i++){
    delete node[i];
    delete weight[i];
  } 
  delete[] node;
  delete[] weight;
  for(i = 0; i <= m_dim; i++) 
    delete[] peano[i];
  delete[] peano;
}



inline void riemann::set_riemann(int k_i, int k, interval tk){
  interval r;
  m_index    = k_i;
  m          = k; 
  r          = peano[m_index][n_index] * tk;
  if (m != 2 * n) r -= r;
  r_inf      = Inf(r);
  r_sup      = Sup(r);
  dremainder = diam(r);
  return;

}


inline void riemann::set_riemann(interval tk, int k_i, int k){
 interval r;
  n_index    = k_i;
  n          = k; 
  r          = peano[m_index][n_index] * 
               power((sup- interval(inf,inf))* 0.5, m + 1) * tk;
  if (m != 2 * n) r  -= r; 
  r_inf      = Inf(r);
  r_sup      = Sup(r);
  dremainder = diam(r);
  return;
}


void riemann::add_quadraturesum(){ 
 

  interval  *x   = node[n_index] , 
            *w   = weight[n_index],
            m_deriv, c;
 
  m_deriv  = (sup - interval(inf,inf)) * 0.5;
  c        = (sup + interval(inf,inf)) * 0.5;

  for(int i=0; i< n; i++)
   accumulate(ri.Quadsum, m_deriv * w[i] , ri.f.value(m_deriv * x[i] + c )); 
}


void riemann::subdivide(){
          
   real eps_local = ri.eps_abs * (sup -inf) / (ri.sup - ri.inf),   
     //  performs better than
     //  eps_local = ri.eps_abs * 0.5 * (sup -inf) / (ri.sup - ri.inf) 

        midpoint;

   r_dot = 0;
   accumulate(r_dot, 0.5,  inf);
   accumulate(r_dot, 0.5,  sup);
   rnd(r_dot, midpoint);
   riemann* left  = new riemann(inf, midpoint, m_index, n_index, ri);
   left->choose_qf(eps_local);
   ri.push(*left);
   inf = midpoint; 
   choose_qf(eps_local);
   quadsum = false;
   ri.push(*this);
}
 
      
void riemann::choose_qf(real eps_local){
   interval tk_old, tk_new;
   real  dr_old, dr_new, factor;  
   int  k, k_index;

   tk_old  = ri.f.taylor(m, interval(inf, sup) ); 
   factor  = pow( (sup-inf) * 0.5, m + 1 ) * diam(tk_old);

   // looking for appropriate n
   dr_old = Sup(peano[m_index][n_dim]) * factor; 
   if (m != 2 * n) dr_old *= 2.0; 

   if (dr_old < eps_local){
    
     for(k_index = n_dim - 1, k = n_max - n_stride; 
         k_index >= 0; k_index--, k -= n_stride)
     {
        dr_new = Sup(peano[m_index][k_index]) * factor;
        if (m != 2 * k) dr_new *= 2.0;
        if(dr_new > eps_local) break;  
     }
     set_riemann(tk_old, k_index + 1, k + n_stride);          
     return;    
   }

   // looking for appropriate m
   if (m_index > 0){  
     k = m - m_stride; k_index = m_index -1;
     tk_new = power( (sup - interval(inf,inf) ) * 0.5, k + 1) * ri.f.taylor(k);
     if ( 2 * diam(tk_new * peano[k_index][n_dim]) < dr_old){ 
       set_riemann(k_index, k, tk_new);          
       return; 
     }
   }

   if(m < m_max){ 
     k = m + m_stride; k_index = m_index + 1;
     tk_new = power( ( sup - interval(inf,inf) ) * 0.5, k + 1) * 
       ri.f.taylor( k );   // remove!!!! interval 
    if (k != 2 * n_max) dr_old *= 0.5; 
    if (diam(tk_new * peano[k_index][n_dim])  < dr_old){ 
       set_riemann(k_index, k, tk_new);  
       return; 
    }
   }
   set_riemann(tk_old, n_index, n);
}


// ----------------------------------------------------------------------------
// --- class sqrt_interval ----------------------------------------------------
// ----------------------------------------------------------------------------


interval** sqrt_interval::init_data(){

  // init node, weight, peano

  node   = new interval*[n_dim + 1];
  weight = new interval*[n_dim + 1];
  peano  = new interval*[n_dim + 1];
  
  for(int i = 0; i<= n_dim; i++){
    node[i]   = new interval[n_min + i * n_stride];
    weight[i] = new interval[n_min + i * n_stride];
  } 

  read_data("./data/gausssqrt6.hex" , node, weight, n_min, n_stride);
  read_data("./data/gausssqrt9.hex", node, weight, n_min, n_stride);
  read_data("./data/gausssqrt12.hex", node, weight, n_min, n_stride);
 
  for(int i = 0; i <= n_dim; i++) 
    peano[i] = new interval[n_dim + 1];
 
  read_data("./data/peanosqrt6.hex", peano ,n_max , n_min, n_stride,
                                                   n_min, n_stride);

  read_data("./data/peanosqrt9.hex", peano, n_max, n_min, n_stride,
                                                   n_min, n_stride);

  read_data("./data/peanosqrt12.hex", peano, n_max, n_min, n_stride,
                                                   n_min, n_stride);
  return peano;
}



void sqrt_interval::free_data(){

 int i;

  for(i = 0; i<= n_dim; i++){
    delete node[i];
    delete weight[i];
  } 
  delete[] node;
  delete[] weight;
  for(i = 0; i <= n_dim; i++) 
    delete[] peano[i];
  delete[] peano;
}



inline void sqrt_interval::set_sqrt(int k_i, int k, interval tk){
  interval r;
  m_index    = k_i;
  m          = k; 
  r          = peano[m_index][n_index] * tk;
  if (m != 2 * n) r -= r;
  r_inf      = Inf(r);
  r_sup      = Sup(r);
  dremainder = diam(r);
  return;

}


inline void sqrt_interval::set_sqrt(interval tk, int k_i, int k){
 interval r;
  n_index    = k_i;
  n          = k; 
  r          = peano[m_index][n_index] * sqrt( interval(sup)) 
               * power( interval(sup), m) * tk;
  if (m != 2 * n) r  -= r; 
  r_inf      = Inf(r);
  r_sup      = Sup(r);
  dremainder = diam(r);
  return;
}


void sqrt_interval::add_quadraturesum(){ 
 

  interval  *x   = node[n_index] , 
            *w   = weight[n_index];

  sqrt_integral& sqi = static_cast<sqrt_integral&>(ri);
 
  for(int i=0; i< n; i++)
   accumulate(sqi.Quadsum,  sqrt(interval(sup)) * w[i] , 
                             sqi.f_w.value(sup * x[i])); 
}


void sqrt_interval::subdivide(){

   sqrt_integral& sqi = static_cast<sqrt_integral&>(ri);
       
   real eps_local = sqi.eps_abs * sup / sqi.sup,   
        midpoint;

   r_dot = 0;
   accumulate(r_dot, 0.5,  inf);
   accumulate(r_dot, 0.5,  sup);
   rnd(r_dot, midpoint);
   riemann* right  = new riemann(midpoint, sup, m_index, n_index, ri);
   right->choose_qf(eps_local);
   sqi.push(*right);
   sup = midpoint; 
   choose_qf(eps_local);
   quadsum = false;
   sqi.push(*this);
}
 
      
void sqrt_interval::choose_qf(const real& eps_local){
   interval tk_old, tk_new;
   real  dr_old, dr_new, factor;  
   int  k, k_index;

   sqrt_integral& sqi = static_cast<sqrt_integral&>(ri);

   tk_old  = sqi.f_w.taylor(m, interval(inf, sup) ); 
   factor  = pow(sup , m + 0.5 ) * diam(tk_old);

   // looking for appropriate n
   dr_old = Sup(peano[m_index][n_dim]) * factor; 
   if (m != 2 * n) dr_old *= 2.0; 

   if (dr_old < eps_local){
    
     for(k_index = n_dim - 1, k = n_max - n_stride; 
         k_index >= 0; k_index--, k -= n_stride)
     {
        dr_new = Sup(peano[m_index][k_index]) * factor;
        if (m != 2 * k) dr_new *= 2.0;
        if(dr_new > eps_local) break;  
     }
     set_sqrt(tk_old, k_index + 1, k + n_stride);          
     return;    
   }

   // looking for appropriate m
   if (m_index > 0){  
     k = m - n_stride; k_index = m_index -1;
     tk_new = power( interval(sup), k ) *
              sqrt( interval(sup)) * sqi.f_w.taylor(k);
     if ( 2 * diam(tk_new * peano[k_index][n_dim]) < dr_old){ 
       set_sqrt(k_index, k, tk_new);          
       return; 
     }
   }

   if(m < n_max){ 
     k = m + n_stride; k_index = m_index + 1;
     tk_new = power( interval(sup), k) * sqrt(interval(sup)) * 
       sqi.f_w.taylor( k );   // remove!!!! interval 
    if (k != 2 * n_max) dr_old *= 0.5; 
    if (diam(tk_new * peano[k_index][n_dim])  < dr_old){ 
       set_sqrt(k_index, k, tk_new);  
       return; 
    }
   }
   set_sqrt(tk_old, n_index, n);
}



// ----------------------------------------------------------------------------
// --- class sqrt_integral ----------------------------------------------------
// ----------------------------------------------------------------------------


sqrt_integral::sqrt_integral(integrand& g):f_w(sqrt_interval::n_max){ 
  operand s;
  f_w = g;
  f  = f_w / sqrt(s);
  // cout << f << endl;
  // cout << g << endl;
  if(!sqrt_counter) sqrt_interval::init_data();
  sqrt_counter++; 
  eps_rel = 0;
}

sqrt_integral::~sqrt_integral(){ 
  sqrt_counter-- ;
  if(!sqrt_counter) sqrt_interval::free_data(); 
}


void sqrt_integral::start_division(){
  sqrt_interval* tmp = new sqrt_interval(0, sup, 1, 1, *this);
  tmp->choose_qf(eps_abs);
  push(*tmp); 
}

// ----------------------------------------------------------------------------
// --- class cauchy_integral --------------------------------------------------
// ----------------------------------------------------------------------------


cauchy_integral::cauchy_integral(): f_c(2*riemann::m_max + 2){
  lambda = 0.0;
}


cauchy_integral::cauchy_integral(integrand& g, real l): f_c(cauchy::m_max){
  operand s;
  f_c  = g;
  lambda  = l;
  f   = f_c / (s - lambda);
  // cout << f << endl;
  // cout << g << endl;
  eps_rel = 0;
}


void cauchy_integral::start_division(){

  int  m_index = (cauchy::m_max - cauchy::m_min) / (cauchy::m_stride),
      rm_index = (riemann::m_max - riemann::m_min) / (riemann::m_stride),
       n_index = (riemann::n_max - riemann::n_min) / (riemann::n_stride);
 
  dotprecision r_dot;

  real p;

  cauchy* c_interval;
  riemann* r_interval;

  r_dot = -inf;
  accumulate(r_dot, -1.0, sup);
  accumulate(r_dot,  2.0, lambda); 
  if ( r_dot == 0.0){  // wow perfect! only one interval!
    c_interval = new cauchy(inf, sup, m_index, n_index, *this);
    c_interval->choose_qf(eps_abs);
    push(*c_interval);
          
  }
  else if ( r_dot >= 0.0){ // lambda right
    r_dot = -sup;
    accumulate(r_dot,  2.0, lambda);
    rnd(r_dot, p);
    if (r_dot != p) throw ;
    r_interval = new riemann(inf, p, rm_index, n_index, *this);
    r_interval->choose_qf(eps_abs * (p - inf) / (sup -inf));
    push(*r_interval);
    c_interval = new cauchy(p, sup, m_index, n_index, *this);
    c_interval->choose_qf(eps_abs * (sup - p) / (sup -inf));
    push(*c_interval);
  }
  else{ // lambda left
    r_dot = -inf;
    accumulate(r_dot,  2.0, lambda);
    rnd(r_dot, p);
    if (r_dot != p) throw ;
    c_interval = new cauchy(inf, p, m_index, n_index, *this);
    c_interval->choose_qf(eps_abs * (p - inf) / (sup -inf) );
    push(*c_interval);
    r_interval = new riemann(p, sup, rm_index, n_index, *this);
    r_interval->choose_qf(eps_abs * (sup - p) / (sup -inf) );
    push(*r_interval);
  }
}



// ----------------------------------------------------------------------------
// --- class cauchy --- -------------------------------------------------------
// ----------------------------------------------------------------------------


void cauchy::add_quadraturesum(){

   interval  *x   = riemann::node[n_index] , 
             *w   = riemann::weight[n_index],
             m, c;
   int i,j;

   //    cauchy_integral& ci = dynamic_cast<cauchy_integral&>(ri);
   cauchy_integral& ci = static_cast<cauchy_integral&>(ri);

   m = (sup - interval(inf,inf)) * 0.5;
   c = (sup + interval(inf,inf)) * 0.5;

   if (n%2){
    j = n / 2;
    accumulate(ci.Quadsum, w[j] * m, ci.f_c.taylor(1,m * x[j] + c));
    for(i=0; i<j; i++) 
      accumulate(ci.Quadsum, w[i] / x[i] , ci.f_c.value(m * x[i] + c )); 
    for(i=j+1; i<n; i++)
      accumulate(ci.Quadsum, w[i] / x[i] , ci.f_c.value(m * x[i] + c )); 
   }
   else for(i=0; i<n; i++)
    accumulate(ci.Quadsum, w[i] / x[i] , ci.f_c.value(m * x[i] + c )); 
}



interval** cauchy::init_data(){
  // init error constants
 
  peano  = new interval*[m_dim + 1];
  for(int i = 0; i <= m_dim; i++) 
    peano[i] = new interval[riemann::n_dim + 1];
 
  read_data("./data/peanocauchy8.hex", peano ,m_max , m_min, m_stride,
                                       riemann::n_min, riemann::n_stride);

  read_data("./data/peanocauchy11.hex", peano, m_max, m_min, m_stride,
                                        riemann::n_min, riemann::n_stride);

  read_data("./data/peanocauchy14.hex", peano, m_max, m_min, m_stride,
                                        riemann::n_min, riemann::n_stride);

  read_data("./data/peanocauchy17.hex", peano, m_max, m_min, m_stride,
                                        riemann::n_min, riemann::n_stride);
  return peano;
}



void cauchy::subdivide(){
  // trisection: three equal parts

   //    cauchy_integral& ci = dynamic_cast<cauchy_integral&>(ri);
   cauchy_integral& ci = static_cast<cauchy_integral&>(ri); 

   real eps_local = ci.eps_abs * (sup -inf) / ( 2.0 * (ci.sup - ci.inf)),
     //  eps_local = ci.eps_abs * (sup -inf) / ( 4.0 * (ci.sup - ci.inf)),
        p1, p2;

   r_dot = 0;
   if (sup >= -inf){    // lambda > 0
     accumulate(r_dot, 0.25,  inf);
     accumulate(r_dot, 0.75,  sup);
     rnd(r_dot, p2);
     r_dot = 0.0; 
     accumulate(r_dot,  1.0,  sup);
     accumulate(r_dot,  1.0,  inf);
     accumulate(r_dot, -1.0,   p2);
     rnd(r_dot, p1); 
     if (r_dot!=p1) throw ;  // no, that can't be true!
   }
   else{
     accumulate(r_dot, 0.75,  inf);
     accumulate(r_dot, 0.25,  sup);
     rnd(r_dot, p1);
     r_dot = 0; 
     accumulate(r_dot,  1.0, sup);
     accumulate(r_dot,  1.0, inf);
     accumulate(r_dot, -1.0,  p1);
     rnd(r_dot, p2); 
     if (r_dot != p2) throw ;  // no, that can't be true!
   }
     
   riemann* left   = new riemann(inf, p1, m_index, n_index, ri);
   left->choose_qf(eps_local);
   ci.push(*left);
   riemann* right = new riemann(p2, sup, m_index, n_index, ri);
   right->choose_qf(eps_local);
   ci.push(*right);
   inf = p1;
   sup = p2; 
   choose_qf( 2 * eps_local);
   quadsum = false;
   ci.push(*this);
}



void cauchy::free_data(){

  for(int i = 0; i <= m_dim; i++) 
    delete[] peano[i];
  delete[] peano;
}


void cauchy::choose_qf(const real& eps_local){

   //    cauchy_integral& ci = dynamic_cast<cauchy_integral&>(ri);
   cauchy_integral& ci = static_cast<cauchy_integral&>(ri); 

   interval tk_old, tk_new;
   real  dr_old, dr_new, factor;  
   int  k, k_index;

   tk_old  = ci.f_c.taylor(m, interval(inf, sup) ); 
   factor  = power( (sup-inf) * 0.5, m ) * diam(tk_old);

   // looking for appropriate n
   
   dr_old = Sup(peano[m_index][riemann::n_dim]) * factor; 
   if (m != 2 * n + 1) dr_old *= 2.0; 

   if (dr_old < eps_local){
    
     for(k_index = riemann::n_dim - 1, k = riemann::n_max - riemann::n_stride; 
         k_index >= 0; k_index--, k -= riemann::n_stride)
     {
        dr_new = Sup(peano[m_index][k_index]) * factor;
        if (m != 2 * k) dr_new *= 2.0;
        if(dr_new > eps_local) break;  
     }
     set_cauchy(tk_old, k_index + 1, k + riemann::n_stride);          
     return;    
   }

   // looking for appropriate m
   if (m_index > 0){  
     k = m - m_stride; k_index = m_index -1;
     tk_new = power( (sup - interval(inf,inf) ) * 0.5, k ) 
                                                * ci.f_c.taylor(k);
     if ( 2 * diam(tk_new * peano[k_index][riemann::n_dim]) < dr_old){ 
       set_cauchy(k_index, k, tk_new);          
       return; 
     }
   }

   if(m < m_max){ 
     k = m + m_stride; k_index = m_index + 1;
     tk_new = power( ( sup - interval(inf,inf) ) * 0.5, k ) * 
              ci.f_c.taylor(k);
    if (k != 2 * riemann::n_max + 1) dr_old *= 0.5; 
    if (diam(tk_new * peano[k_index][riemann::n_dim])  < dr_old){ 
       set_cauchy(k_index, k, tk_new);  
       return; 
    }
   }
   set_cauchy(tk_old, n_index, n);
}


void cauchy::set_cauchy(int k_i, int k, interval tk){
  interval r;
  m_index    = k_i;
  m          = k; 
  r          = peano[m_index][n_index] * tk;
  if (m != 2 * n + 1) r -= r;
  r_inf      = Inf(r);
  r_sup      = Sup(r);
  dremainder = diam(r);
  return;
}


void cauchy::set_cauchy(interval tk, int k_i, int k){
  interval r;
  n_index    = k_i;
  n          = k; 
  r          = peano[m_index][n_index] * 
               power((sup- interval(inf,inf))* 0.5, m) * tk;
  if (m != 2 * n + 1) r  -= r; 
  r_inf      = Inf(r);
  r_sup      = Sup(r);
  dremainder = diam(r);
  return;
}


// ----------------------------------------------------------------------------
// --- class hadamard_integral ------------------------------------------------
// ----------------------------------------------------------------------------

hadamard_integral::hadamard_integral(integrand& g, real l): cauchy_integral()
{
  operand s;
  f_c  = g;
  lambda  = l;
  f   = f_c / sqr(s - lambda);
  // cout << f << endl;
  // cout << g << endl;
  eps_rel = 0;
}


void hadamard_integral::start_division(){

  int m_index = (riemann::m_max - riemann::m_min) / (riemann::m_stride),
      n_index = (riemann::n_max - riemann::n_min) / (riemann::n_stride);
 
  dotprecision r_dot;

  real p;

  hadamard* h_interval;
  riemann*  r_interval;

  r_dot = -inf;
  accumulate(r_dot, -1.0, sup);
  accumulate(r_dot,  2.0, lambda); 
  if ( r_dot == 0.0){  // lambda is the midpoint
    h_interval = new hadamard(inf, sup, *this);
    h_interval->choose_qf(eps_abs);
    push(*h_interval);
          
  }
  else if ( r_dot >= 0.0){ // lambda right
    r_dot = -sup;
    accumulate(r_dot,  2.0, lambda);
    rnd(r_dot, p);
    if (r_dot != p) throw ;
    r_interval = new riemann(inf, p, m_index, n_index, *this);
    r_interval->choose_qf(eps_abs * (p - inf) / (sup -inf));
    push(*r_interval);
    h_interval = new hadamard(p, sup, *this);
    h_interval->choose_qf(eps_abs * (sup - p) / (sup -inf));
    push(*h_interval);
  }
  else{ // lambda left
    r_dot = -inf;
    accumulate(r_dot,  2.0, lambda);
    rnd(r_dot, p);
    if (r_dot != p) throw ;
    h_interval = new hadamard(inf, p,  *this);
    h_interval->choose_qf(eps_abs * (p - inf) / (sup -inf) );
    push(*h_interval);
    r_interval = new riemann(p, sup, m_index, n_index, *this);
    r_interval->choose_qf(eps_abs * (sup - p) / (sup -inf) );
    push(*r_interval);
  }
}


// ----------------------------------------------------------------------------
// --- class hadamard -------- ------------------------------------------------
// ----------------------------------------------------------------------------

interval* hadamard::init_data(){
// n = 8,11,14,17 m = 18,24,30,36
  peano = new interval[had_max + 1];

  peano[0] =   interval(131072,131072) / (17.0*sqr(interval(12870,12870)));
  peano[1] =   interval(131072,131072) / 
               ( 23.0 * sqr( interval(6783,6783) * 13 ) );      
  peano[2] =   2 * sqr( interval(2048) ) /
               ( 29.0 * sqr( interval(15525,15525)*323 ) );
  peano[3] =   2.0*sqr(sqr(interval(256,256)))/
               (3500 * sqr(24273 * interval( 4807 ) )); 

  return peano;
}


void hadamard::set_hadamard(interval r, int k, int l){
  interval remainder;
  m          = k;
  n          = l;
  remainder  =  r; 
  r_inf      = Inf(remainder);
  r_sup      = Sup(remainder);
  dremainder = diam(remainder);
  return;
}



void hadamard::choose_qf(const real& eps_local){
   // hadamard_integral& ci = dynamic_cast<hadamard_integral&>(ri);
   hadamard_integral& hi = static_cast<hadamard_integral&>(ri); 

   interval r_old, r_new;
   int  k, k_dim;

   m = 2 * n + 2; m_index = n_index;
   r_old = hi.f_c.taylor(m, interval(inf, sup)) * 
           power( (sup - inf) / 2, m - 1) * peano[m_index];

   if (m_index){
      k_dim = m_index - 2;
      k     = 2 *  (n - 2 * riemann::n_stride) + 2;
      r_new = hi.f_c.taylor(k, interval(inf, sup)) * 
              power( ( sup - inf ) / 2, k - 1) * peano[k_dim];
     if (diam(r_new)< diam(r_old)){
       n_index -=2;
       m_index -=2;
       set_hadamard(r_new, k, n - 2*riemann::n_stride);
       return;
     }
   }
   if (m_index == m_dim){
      k_dim = m_index + 2;
      k     = 2 *  (n + 2 * riemann::n_stride) + 2;
      r_new = hi.f_c.taylor(k, interval(inf, sup)) * 
              power( ( sup - inf ) / 2, k - 1) * peano[k_dim];
      if (diam(r_new) < diam(r_old)){
       n_index += 2;
       m_index += 2;
       set_hadamard(r_new, k, n + 2 * riemann::n_stride);
       return;
     }  
   }
   set_hadamard(r_old, m, n);
}



void hadamard::add_quadraturesum(){

   interval  *x   = riemann::node[n_index] , 
             *w   = riemann::weight[n_index],
             m, c,  p, q, qq, help;
   int i;

   //    hadamard_integral& ci = dynamic_cast<hadamard_integral&>(ri);
   hadamard_integral& hi = static_cast<hadamard_integral&>(ri);

   m = (sup - interval(inf,inf)) * 0.5;
   c = (sup + interval(inf,inf)) * 0.5;

   for(i=0; i < n; i++) 
    accumulate(hi.Quadsum, w[i] / (sqr(x[i]) * m ) , 
                           hi.f_c.value(m * x[i] + c )); 
    
    p  = -interval(0.5,0.5);
    q  =  interval(2,2);
    qq =  interval(4,4);

     for(i = 2; i <= n /2; i++){
       help = 2 * i;
     p  *= - (help - 1) / help;
     q  *= - (help - 2) / (help -1);
     qq  = - (help - 1) / help * qq + ( (2 * help) - 1 ) / help * q; 
     }
     qq /= p;

     accumulate(hi.Quadsum, qq / m , hi.f_c.value(c) );
}

operand x, y(operand::Var_2);
